# Copyright (c) 2023 Advanced Micro Devices, Inc.
# All rights reserved.
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:
#
# * Redistributions of source code must retain the above copyright notice, this
#   list of conditions and the following disclaimer.
#
# * Redistributions in binary form must reproduce the above copyright notice,
#   this list of conditions and the following disclaimer in the documentation
#   and/or other materials provided with the distribution.
#
# * Neither the name of qonnx nor the names of its
#   contributors may be used to endorse or promote products derived from
#   this software without specific prior written permission.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
# AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
# IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
# DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE
# FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
# DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
# SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER
# CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,
# OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
# OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

import clize
import itertools
import numpy as np
import pprint
from warnings import warn

from qonnx.core.modelwrapper import ModelWrapper
from qonnx.core.onnx_exec import execute_node
from qonnx.transformation.fold_constants import FoldConstants
from qonnx.transformation.infer_datatypes import InferDataTypes
from qonnx.transformation.infer_shapes import InferShapes
from qonnx.util.basic import get_by_name
from qonnx.util.cleanup import cleanup_model
from qonnx.util.onnx import valueinfo_to_tensor

# walk the graph to deduce range information about each tensor
# assumptions:
# - layout and shape inference already completed
# - any quantized weights are resolved into initializers
# - range info is generated per-channel (tuple of 1D arrays) or per-tensor (tuple of scalars)


def calculate_matvec_accumulator_extremum(matrix: np.ndarray, vec_min, vec_max):
    """Calculate the minimum and maximum possible result (accumulator) values for a dot product A*x,
    given matrix A of dims (MH, MW), and vector (MW) with range (vec_min, vec_max). vec_min and
    vec_max are either scalars, or 1D arrays of length MW.
    Returns (acc_min, acc_max) where acc_min and acc_max are 1D arrays of length MH."""
    max_vectors = np.where(matrix > 0, vec_max, vec_min)
    min_vectors = np.where(matrix > 0, vec_min, vec_max)
    max_values = (matrix * max_vectors).sum(axis=1)
    min_values = (matrix * min_vectors).sum(axis=1)
    return (min_values, max_values)


def calc_gemm_range(node, model, range_dict):
    alpha = get_by_name(node.attribute, "alpha").f
    beta = get_by_name(node.attribute, "beta").f
    transA = get_by_name(node.attribute, "transA")
    if transA is not None:
        transA = transA.i
    else:
        transA = 0
    transB = get_by_name(node.attribute, "transB")
    if transB is not None:
        transB = transB.i
    else:
        transB = 1
    assert (not transA) and transB
    iname = node.input[0]
    wname = node.input[1]
    bname = None
    if len(node.input) > 2:
        bname = node.input[2]
    oname = node.output[0]

    irange = range_dict[iname]
    imin, imax = irange
    weights = model.get_initializer(wname)
    assert weights is not None, "Uninitialized Gemm weights"
    if type(imin) is np.ndarray:
        assert len(imin) == weights.shape[1], "Dot product length mismatch, np broadcast may be wrong"
    pmin, pmax = calculate_matvec_accumulator_extremum(weights, imin, imax)
    # apply Gemm scale factors to matrix multiply output
    pmin *= alpha
    pmax *= alpha
    # if there is a bias, apply it to the range
    if bname is not None:
        bias = model.get_initializer(bname)
        assert bias is not None, "Uninitialized Gemm bias"
        pmin += beta * bias
        pmax += beta * bias
    ret = (pmin, pmax)
    range_dict[oname] = ret


def calc_matmul_range(node, model, range_dict):
    iname = node.input[0]
    wname = node.input[1]
    oname = node.output[0]
    irange = range_dict[iname]
    imin, imax = irange
    weights = model.get_initializer(wname)
    assert weights is not None, "Uninitialized MatMul weights"
    # util function expects (mh, mw) so transpose
    weights = weights.transpose()
    if type(imin) is np.ndarray:
        assert len(imin) == weights.shape[1], "Dot product length mismatch, np broadcast may be wrong"
    ret = calculate_matvec_accumulator_extremum(weights, imin, imax)
    range_dict[oname] = ret


def calc_conv_range(node, model, range_dict):
    iname = node.input[0]
    wname = node.input[1]
    assert len(node.input) == 2, "Found unsupported Conv with bias"
    oname = node.output[0]
    irange = range_dict[iname]
    imin, imax = irange
    weights = model.get_initializer(wname)
    assert weights is not None, "Uninitialized Conv weights"
    # do weight reshaping to treat Conv similar to MatMul
    # (mh, mw) = (ofm, (ifm x k0 x k1 x ...))
    conv_ofm = weights.shape[0]
    conv_ifm = weights.shape[1]
    weights = weights.reshape(conv_ofm, -1)
    k_total = weights.shape[1] // conv_ifm
    groups = get_by_name(node.attribute, "group")
    if groups is None:
        # default to dense convs
        groups = 1
    else:
        groups = groups.i
    # TODO smarter check, other kinds of grouped convs out there..
    is_depthwise = groups > 1
    # need to construct specialzed input range vectors for Conv
    if is_depthwise:
        conv_ifm = conv_ofm
    if type(imin) is np.ndarray:
        imin_rep = np.repeat(imin, k_total)
        imax_rep = np.repeat(imax, k_total)
    else:
        imin_rep = imin
        imax_rep = imax
    dw_ret_min = []
    dw_ret_max = []
    for i in range(conv_ofm):
        w_slice = weights[i, :].reshape(1, -1)
        if is_depthwise and type(imin_rep) is np.ndarray:
            dw_ret = calculate_matvec_accumulator_extremum(
                w_slice, imin_rep[i * k_total : (i + 1) * k_total], imax_rep[i * k_total : (i + 1) * k_total]
            )
        else:
            dw_ret = calculate_matvec_accumulator_extremum(w_slice, imin_rep, imax_rep)
        dw_ret_min.append(dw_ret[0].item())
        dw_ret_max.append(dw_ret[1].item())
    ret = (np.asarray(dw_ret_min), np.asarray(dw_ret_max))
    range_dict[oname] = ret


def calc_convtranspose_range(node, model, range_dict):
    iname = node.input[0]
    wname = node.input[1]
    assert len(node.input) == 2, "Found unsupported ConvTranspose with bias"
    oname = node.output[0]
    irange = range_dict[iname]
    imin, imax = irange
    weights = model.get_initializer(wname)
    assert weights is not None, "Uninitialized ConvTranspose weights"
    groups = get_by_name(node.attribute, "group")
    if groups is None:
        # default to dense convs
        groups = 1
    else:
        groups = groups.i
    assert groups == 1, "Only dense (non-grouped) ConvTranspose is supported"
    # do weight reshaping to treat Conv similar to MatMul
    # (mh, mw) = (ofm, (ifm x k0 x k1 x ...))
    conv_ofm = weights.shape[1]
    conv_ifm = weights.shape[0]
    weights = weights.transpose(1, 0, 2, 3).reshape(conv_ofm, -1)
    k_total = weights.shape[1] // conv_ifm
    if type(imin) is np.ndarray:
        imin_rep = np.repeat(imin, k_total)
        imax_rep = np.repeat(imax, k_total)
    else:
        imin_rep = imin
        imax_rep = imax
    dw_ret_min = []
    dw_ret_max = []
    for i in range(conv_ofm):
        w_slice = weights[i, :].reshape(1, -1)
        dw_ret = calculate_matvec_accumulator_extremum(w_slice, imin_rep, imax_rep)
        dw_ret_min.append(dw_ret[0].item())
        dw_ret_max.append(dw_ret[1].item())
    ret = (np.asarray(dw_ret_min), np.asarray(dw_ret_max))
    range_dict[oname] = ret


def get_minmax_prototype_tensors(irange, ishp, inp_vi, i_channel_axis=1):
    proto_min = valueinfo_to_tensor(inp_vi)
    proto_max = valueinfo_to_tensor(inp_vi)
    if type(irange[0]) in [float, int, np.float16, np.float32, np.float64, np.uint8, np.int8]:
        imin, imax = irange
        proto_min[...] = imin
        proto_max[...] = imax
    elif type(irange[0]) is np.ndarray:
        # irange is [(min_ch0, max_ch0), (min_ch1, max_ch1) ...]
        n_ch = ishp[i_channel_axis]
        proto_min = np.moveaxis(proto_min, i_channel_axis, 0)
        proto_max = np.moveaxis(proto_max, i_channel_axis, 0)
        for ch in range(n_ch):
            proto_min[ch, ...] = irange[0][ch]
            proto_max[ch, ...] = irange[1][ch]
        proto_min = np.moveaxis(proto_min, 0, i_channel_axis)
        proto_max = np.moveaxis(proto_max, 0, i_channel_axis)
    else:
        assert False, "Unknown range type"
    return (proto_min, proto_max)


def is_dyn_input(x, model):
    return model.get_initializer(x) is None and x != ""


def calc_monotonic_range(node, model, range_dict, i_channel_axis=1):
    opset_version = model.model.opset_import[0].version
    oname = node.output[0]
    dyn_inps = [x for x in node.input if is_dyn_input(x, model)]
    n_dyn_inp = len(dyn_inps)
    proto_vectors = []
    # generate min-max prototype vectors for each dynamic input
    for inp in dyn_inps:
        irange = range_dict[inp]
        ishp = model.get_tensor_shape(inp)
        inp_vi = model.get_tensor_valueinfo(inp)
        proto_vectors.append(get_minmax_prototype_tensors(irange, ishp, inp_vi, i_channel_axis))
    # process all combinations of prototype vectors for dynamic inputs
    running_min = [None for i in range(len(node.output))]
    running_max = [None for i in range(len(node.output))]
    # create context for single-node execution
    ctx = {x: model.get_initializer(x) for x in node.input}
    for oname in node.output:
        ctx[oname] = valueinfo_to_tensor(model.get_tensor_valueinfo(oname))
    # assume all outputs are homogenous wrt data layout (e.g. channel axis
    # always lives in the same position)
    axes_to_min = [i for i in range(ctx[oname].ndim)]
    axes_to_min.remove(i_channel_axis)
    axes_to_min = tuple(axes_to_min)
    for inps in itertools.product(*proto_vectors):
        for i in range(n_dyn_inp):
            ctx[dyn_inps[i]] = inps[i]
        execute_node(node, ctx, model.graph, opset_version=opset_version)
        for oind, oname in enumerate(node.output):
            # grab new output and update running min/max
            out = ctx[oname]
            chanwise_min = out.min(axis=axes_to_min).flatten()
            chanwise_max = out.max(axis=axes_to_min).flatten()
            running_min[oind] = (
                np.minimum(chanwise_min, running_min[oind]).flatten() if running_min[oind] is not None else chanwise_min
            )
            running_max[oind] = (
                np.maximum(chanwise_max, running_max[oind]).flatten() if running_max[oind] is not None else chanwise_max
            )
    for oind, oname in enumerate(node.output):
        range_dict[oname] = (running_min[oind], running_max[oind])


def calc_range_outdtype(node, model, range_dict):
    oname = node.output[0]
    odt = model.get_tensor_datatype(oname)
    assert odt is not None, "Cannot infer %s range, dtype annotation is missing" % oname
    range_dict[oname] = (odt.min(), odt.max())


optype_to_range_calc = {
    "Transpose": calc_monotonic_range,
    "MatMul": calc_matmul_range,
    "Conv": calc_conv_range,
    "ConvTranspose": calc_convtranspose_range,
    "QuantMaxNorm": calc_range_outdtype,
    "Flatten": calc_monotonic_range,
    "Reshape": calc_monotonic_range,
    "Quant": calc_monotonic_range,
    "BipolarQuant": calc_monotonic_range,
    "Mul": calc_monotonic_range,
    "Sub": calc_monotonic_range,
    "Div": calc_monotonic_range,
    "Add": calc_monotonic_range,
    "BatchNormalization": calc_monotonic_range,
    "Relu": calc_monotonic_range,
    "Pad": calc_monotonic_range,
    "AveragePool": calc_monotonic_range,
    "Trunc": calc_range_outdtype,
    "MaxPool": calc_monotonic_range,
    "Resize": calc_monotonic_range,
    "Upsample": calc_monotonic_range,
    "GlobalAveragePool": calc_monotonic_range,
    "Gemm": calc_gemm_range,
    "QuantizeLinear": calc_monotonic_range,
    "DequantizeLinear": calc_monotonic_range,
    "Clip": calc_monotonic_range,
    "Sigmoid": calc_monotonic_range,
    "Concat": calc_monotonic_range,
    "Split": calc_monotonic_range,
}


def simplify_range(range):
    """Where possible, simplify a range that is expressed as channelwise ranges
    back to a scalar range if all channels' ranges were equal."""
    rmin = range[0]
    rmax = range[1]
    if type(rmin) is np.ndarray and type(rmax) is np.ndarray:
        rmin_eq = all(rmin == rmin[0])
        rmax_eq = all(rmax == rmax[0])
        if rmin_eq and rmax_eq:
            return (rmin[0], rmax[0])
        else:
            return range
    else:
        return range


REPORT_MODE_RANGE = "range"
REPORT_MODE_STUCKCHANNEL = "stuck_channel"
REPORT_MODE_ZEROSTUCKCHANNEL = "zerostuck_channel"

report_modes = {REPORT_MODE_RANGE, REPORT_MODE_STUCKCHANNEL, REPORT_MODE_ZEROSTUCKCHANNEL}

report_mode_options = clize.parameters.mapped(
    [
        (REPORT_MODE_RANGE, [REPORT_MODE_RANGE], "Report ranges"),
        (REPORT_MODE_STUCKCHANNEL, [REPORT_MODE_STUCKCHANNEL], "Report stuck channels"),
        (REPORT_MODE_ZEROSTUCKCHANNEL, [REPORT_MODE_ZEROSTUCKCHANNEL], "Report 0-stuck channels"),
    ]
)


def range_analysis(
    model_filename_or_wrapper,
    *,
    irange="",
    key_filter: str = "",
    report_mode: report_mode_options = REPORT_MODE_STUCKCHANNEL,
    prettyprint=False,
    do_cleanup=False
):
    assert report_mode in report_modes, "Unrecognized report_mode, must be " + str(report_modes)
    if isinstance(model_filename_or_wrapper, ModelWrapper):
        model = model_filename_or_wrapper
    else:
        model = ModelWrapper(model_filename_or_wrapper)
    if isinstance(irange, str):
        if irange == "":
            range_min = None
            range_max = None
        else:
            irange = eval(irange)
            range_min, range_max = irange
            if isinstance(range_min, list):
                range_min = np.asarray(range_min, dtype=np.float32)
            if isinstance(range_max, list):
                range_max = np.asarray(range_max, dtype=np.float32)
    elif isinstance(irange, tuple):
        range_min, range_max = irange
    else:
        assert False, "Unknown irange type"
    if do_cleanup:
        model = cleanup_model(model)
    # call constant folding with no exclusions to get weight initializers
    # (but not full cleanup, in order to preserve node/tensor naming)
    model = model.transform(InferShapes())
    model = model.transform(FoldConstants(exclude_op_types=[]))
    model = model.transform(InferDataTypes())
    range_dict = {}
    stuck_chans = {}

    # start by calculating/annotating range info for input tensors
    for inp in model.graph.input:
        iname = inp.name
        if range_min is None or range_max is None:
            # use idt annotation
            idt = model.get_tensor_datatype(iname)
            assert idt is not None, "Could not infer irange, please specify"
            range_min = idt.min()
            range_max = idt.max()
        range_dict[iname] = (range_min, range_max)

    for node in model.graph.node:
        dyn_inputs = [x for x in node.input if is_dyn_input(x, model)]
        inprange_ok = all([x in range_dict.keys() for x in dyn_inputs])
        op_ok = node.op_type in optype_to_range_calc.keys()
        if inprange_ok and op_ok:
            range_calc_fxn = optype_to_range_calc[node.op_type]
            range_calc_fxn(node, model, range_dict)
            out_range = range_dict[node.output[0]]
            tensor_stuck_chans = np.nonzero(out_range[0] == out_range[1])[0]
            if len(tensor_stuck_chans) > 0:
                list_stuck_chans = list(tensor_stuck_chans)
                list_stuck_values = list(out_range[0][tensor_stuck_chans])
                stuck_chans[node.output[0]] = list(zip(list_stuck_chans, list_stuck_values))
            range_dict[node.output[0]] = simplify_range(out_range)
        else:
            warn("Skipping %s : inp_range? %s op_ok? (%s) %s" % (node.name, str(inprange_ok), node.op_type, str(op_ok)))

    # range dict is now complete, apply filters and formatting
    if report_mode in [REPORT_MODE_ZEROSTUCKCHANNEL, REPORT_MODE_STUCKCHANNEL]:
        ret = stuck_chans
    else:
        ret = range_dict
    # only keep tensors (keys) where filter appears in the name
    if key_filter != "":
        ret = {k: v for (k, v) in ret.items() if key_filter in k}

    if report_mode == REPORT_MODE_RANGE:
        # convert ranges in report to regular Python lists
        for tname, trange in ret.items():
            if type(trange[0]) is np.ndarray:
                ret[tname] = (list(trange[0]), list(trange[1]))
    elif report_mode == REPORT_MODE_ZEROSTUCKCHANNEL:
        # only leave channels that are stuck at zero
        # value info removed since implicitly 0
        new_ret = {}
        for tname, schans in ret.items():
            schans_only_zero = set([x[0] for x in schans if x[1] == 0])
            if len(schans_only_zero) > 0:
                new_ret[tname] = schans_only_zero
        ret = new_ret
    if prettyprint:
        ret = pprint.pformat(ret, sort_dicts=False)
    return ret


def main():
    clize.run(range_analysis)


if __name__ == "__main__":
    main()



# Copyright (c) 2020 Xilinx, Inc.
# All rights reserved.
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:
#
# * Redistributions of source code must retain the above copyright notice, this
#   list of conditions and the following disclaimer.
#
# * Redistributions in binary form must reproduce the above copyright notice,
#   this list of conditions and the following disclaimer in the documentation
#   and/or other materials provided with the distribution.
#
# * Neither the name of Xilinx nor the names of its
#   contributors may be used to endorse or promote products derived from
#   this software without specific prior written permission.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
# AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
# IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
# DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE
# FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
# DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
# SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER
# CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,
# OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
# OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

import numpy as np
import warnings
from onnx import TensorProto, helper

from qonnx.transformation.base import Transformation
from qonnx.transformation.extract_conv_bias import ExtractBiasFromConv
from qonnx.util.basic import auto_pad_to_explicit_padding, get_by_name


class LowerConvsToMatMul(Transformation):
    """Replace Conv layers with pairs of Im2Col-MatMul layers, plus Transpose
    layers to keep the original data layout."""

    def apply(self, model):
        model = model.transform(ExtractBiasFromConv())
        graph = model.graph
        graph_modified = False
        for node_ind, node in enumerate(graph.node, start=1):
            if node.op_type != "Conv":
                continue

            if len(node.input) == 3:
                warnings.warn("Found Conv node with bias, skipping")
                continue

            # extract parameters of node
            (
                cnv_input,
                cnv_output,
                cnv_input_datatype,
                cnv_output_datatype,
                k_h,
                k_w,
                stride_h,
                stride_w,
                group,
                weight_name,
                conv_weight_inp_name,
                conv_weight_q_scale_name,
                W_conv,
                ifm_ch,
                ofm_ch,
                ifm_dim_h,
                ifm_dim_w,
                ofm_dim_h,
                ofm_dim_w,
                dilation,
                pad,
            ) = self.extract_conv_params(model, node)

            if W_conv is None:
                warnings.warn("Found Conv node with non-initialized weight, skipping")
                continue

            # if depthwise conv create sparse matrix and variable "dw"
            # to store as attribute in Im2Col that indicates that the created
            # Im2Col node belongs to a depthwise convolution
            dw = False
            if group == ifm_ch and ofm_ch == ifm_ch:
                W_sparse = np.zeros((ofm_ch, ifm_ch, k_h, k_w))  # (OFM, IFM, k_H, k_W)
                # TODO: if the convolution is quantized with a non-zero zeropoint we
                # should be using the zeropoint value here instead of np.zeros
                for ch in range(ifm_ch):
                    W_sparse[ch][ch] = W_conv[ch][0]  # W_conv = [OFM, IFM, k_H, k_W]
                W_conv = W_sparse.astype(np.float32)
                # we need to store information of the
                # sparsity of the weight matrix. For this
                # we use the sparsity annotation of the
                # weight tensor
                sparsity = {"dw": {"kernel_shape": [k_h, k_w]}}
                model.set_tensor_sparsity(weight_name, sparsity)
                # additionally create variable "dw" to store
                # as attribute in Im2Col that indicates that the created
                # Im2Col node belongs to a depthwise convolution
                dw = True

            # reuse conv weights for new matmul weights
            # conv weights are [OFM][IFM][k][k]
            # first convert to [OFM][k_h][k_w][IFM] (to remain compatible with
            # finn-hlslib and how it does im2col/sliding window)
            W_matmul = W_conv.transpose(0, 2, 3, 1)  # W_conv = [OFM, IFM, k_H, k_W]
            # reshape into [OFM][k_h*k_w*IFM] matrix
            W_matmul = W_matmul.reshape(ofm_ch, ifm_ch * k_h * k_w)
            # transpose to get ONNX-compatible [k_h*k_w*IFM][OFM] matrix
            W_matmul = W_matmul.T
            model.set_initializer(weight_name, W_matmul)
            if weight_name != conv_weight_inp_name:
                # required for convs with quantized weights
                model.set_tensor_shape(conv_weight_inp_name, W_matmul.shape)
            if conv_weight_q_scale_name is not None:
                # required for convs with quantized weights
                scale_weight_q = model.get_initializer(conv_weight_q_scale_name)
                if scale_weight_q.ndim > 0:
                    # scale shape is originally [OFM, IFM, k_H, k_W]
                    # transpose into [OFM, k_H, k_W, IFM]
                    scale_weight_q = scale_weight_q.transpose(0, 2, 3, 1)
                    # reshape into [OFM][k_h*k_w*IFM] matrix
                    scale_weight_q = scale_weight_q.reshape(ofm_ch, -1)
                    # transpose to be shape-compatible with weight matrix
                    scale_weight_q = scale_weight_q.T
                    model.set_initializer(conv_weight_q_scale_name, scale_weight_q)

            # create new intermediate values
            inp_trans_out = helper.make_tensor_value_info(
                model.make_new_valueinfo_name(),
                TensorProto.FLOAT,
                (1, ifm_dim_h, ifm_dim_w, ifm_ch),  # NHWC
            )
            graph.value_info.append(inp_trans_out)
            inp_trans_out = inp_trans_out.name
            model.set_tensor_datatype(inp_trans_out, cnv_input_datatype)

            # k_h=k_w==1: pointwise convolution, thus no im2col needed
            need_im2col = any(p != 0 for p in pad) or k_h != 1 or k_w != 1 or stride_h != 1 or stride_w != 1

            # create new intermediate values
            matmul_out = helper.make_tensor_value_info(
                model.make_new_valueinfo_name(), TensorProto.FLOAT, (1, ofm_dim_h, ofm_dim_w, ofm_ch)
            )
            graph.value_info.append(matmul_out)
            matmul_out = matmul_out.name
            model.set_tensor_datatype(matmul_out, cnv_output_datatype)

            # create new nodes
            # NCHW -> NHWC
            inp_trans_node = helper.make_node("Transpose", [cnv_input], [inp_trans_out], perm=[0, 2, 3, 1])
            nodes_to_insert = [inp_trans_node]

            if need_im2col:
                im2col_out = helper.make_tensor_value_info(
                    model.make_new_valueinfo_name(), TensorProto.FLOAT, (1, ofm_dim_h, ofm_dim_w, ifm_ch * k_h * k_w)
                )
                graph.value_info.append(im2col_out)
                im2col_out = im2col_out.name
                model.set_tensor_datatype(im2col_out, cnv_input_datatype)
                im2col_node = helper.make_node(
                    "Im2Col",
                    [inp_trans_out],
                    [im2col_out],
                    domain="qonnx.custom_op.general",
                    stride=[stride_h, stride_w],
                    kernel_size=[k_h, k_w],
                    pad_amount=pad,
                    input_shape="(1,{},{},{})".format(ifm_dim_h, ifm_dim_w, ifm_ch),
                    depthwise=dw,
                    dilations=dilation,
                )
                nodes_to_insert.append(im2col_node)

            matmul_input = im2col_out if need_im2col else inp_trans_out
            # do matmul
            matmul_node = helper.make_node("MatMul", [matmul_input, conv_weight_inp_name], [matmul_out])
            # NHWC -> NCHW
            out_trans_node = helper.make_node("Transpose", [matmul_out], [cnv_output], perm=[0, 3, 1, 2])

            nodes_to_insert.extend([matmul_node, out_trans_node])

            # insert nodes where the conv is to preserve topological ordering
            for i, insert_node in enumerate(nodes_to_insert):
                graph.node.insert(node_ind + i, insert_node)
            graph.node.remove(node)

        return (model, graph_modified)

    def extract_conv_params(self, model, node):
        cnv_input = node.input[0]
        cnv_output = node.output[0]
        cnv_input_datatype = model.get_tensor_datatype(cnv_input)
        cnv_output_datatype = model.get_tensor_datatype(cnv_output)
        k_h = get_by_name(node.attribute, "kernel_shape").ints[0]
        k_w = get_by_name(node.attribute, "kernel_shape").ints[1]
        stride_h = get_by_name(node.attribute, "strides").ints[0]
        stride_w = get_by_name(node.attribute, "strides").ints[1]
        group = get_by_name(node.attribute, "group").i
        weight_name = node.input[1]
        conv_weight_inp_name = node.input[1]
        conv_weight_q_scale_name = None
        W_conv = model.get_initializer(weight_name)
        if W_conv is None:
            # check to see if there is an immediate quantizer node feeding the weight input
            w_producer = model.find_producer(weight_name)
            if not (w_producer is None) and w_producer.op_type == "Quant":
                W_conv = model.get_initializer(w_producer.input[0])
                weight_name = w_producer.input[0]
                conv_weight_q_scale_name = w_producer.input[1]
        ifm_ch = model.get_tensor_shape(cnv_input)[1]  # assume NCHW
        ofm_ch = model.get_tensor_shape(cnv_output)[1]  # assume NCHW
        ifm_dim_h = model.get_tensor_shape(cnv_input)[2]  # assume NCHW
        ifm_dim_w = model.get_tensor_shape(cnv_input)[3]  # assume NCHW
        ofm_dim_h = model.get_tensor_shape(cnv_output)[2]  # assume NCHW
        ofm_dim_w = model.get_tensor_shape(cnv_output)[3]  # assume NCHW
        dilation_attr = get_by_name(node.attribute, "dilations")
        dilation = dilation_attr.ints if dilation_attr is not None else [1, 1]  # default value
        auto_pad = get_by_name(node.attribute, "auto_pad")
        if auto_pad is not None:
            auto_pad = auto_pad.s.decode("utf-8")
            if auto_pad == "NOTSET":
                pad = get_by_name(node.attribute, "pads").ints
            else:
                pad = auto_pad_to_explicit_padding(
                    auto_pad, ifm_dim_h, ifm_dim_w, k_h, k_w, stride_h, stride_w, len(model.get_tensor_shape(cnv_input)) - 2
                )
        else:
            pad = get_by_name(node.attribute, "pads").ints

        if len(pad) == 2:  # only one dimension should be padded
            assert ifm_dim_h == 1 or ifm_dim_w == 1, "Padding is assumed to be 1D, image is 2D"

        return (
            cnv_input,
            cnv_output,
            cnv_input_datatype,
            cnv_output_datatype,
            k_h,
            k_w,
            stride_h,
            stride_w,
            group,
            weight_name,
            conv_weight_inp_name,
            conv_weight_q_scale_name,
            W_conv,
            ifm_ch,
            ofm_ch,
            ifm_dim_h,
            ifm_dim_w,
            ofm_dim_h,
            ofm_dim_w,
            dilation,
            pad,
        )



# Copyright (c) 2023 Advanced Micro Devices, Inc.
# All rights reserved.
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:
#
# * Redistributions of source code must retain the above copyright notice, this
#   list of conditions and the following disclaimer.
#
# * Redistributions in binary form must reproduce the above copyright notice,
#   this list of conditions and the following disclaimer in the documentation
#   and/or other materials provided with the distribution.
#
# * Neither the name of Xilinx nor the names of its
#   contributors may be used to endorse or promote products derived from
#   this software without specific prior written permission.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
# AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
# IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
# DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE
# FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
# DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
# SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER
# CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,
# OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
# OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.


from pkgutil import get_data

import qonnx.util.inference_cost as infc
from qonnx.core.modelwrapper import ModelWrapper
from qonnx.util.cleanup import cleanup_model


def test_matmul_mac_cost():
    raw_model = get_data("qonnx", "data/onnx/matmul_update/sdp.onnx")
    model = ModelWrapper(raw_model)
    cleaned_model = cleanup_model(model)
    # Two Matmul layers with shape (i_shape, w_shape, o_shape),
    # L1: ([4, 64, 32], [4, 32, 64], [4, 64, 64]) and L2: ([4, 64, 64], [4, 64, 32], [4, 64, 32])
    inf_cost_dict = infc.inference_cost(cleaned_model, discount_sparsity=False)["total_cost"]
    mac_cost = inf_cost_dict["op_mac_FLOAT32_FLOAT32"]  # Expected mac cost 4*32*64*64 + 4*64*64*32 = 1048576
    assert mac_cost == 1048576.0, "Error: discrepancy in mac cost."



# Copyright (c) 2022 Advanced Micro Devices, Inc.
# Copyright (c) 2021 Xilinx, Inc.
# All rights reserved.
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:
#
# * Redistributions of source code must retain the above copyright notice, this
#   list of conditions and the following disclaimer.
#
# * Redistributions in binary form must reproduce the above copyright notice,
#   this list of conditions and the following disclaimer in the documentation
#   and/or other materials provided with the distribution.
#
# * Neither the name of Xilinx nor the names of its
#   contributors may be used to endorse or promote products derived from
#   this software without specific prior written permission.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
# AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
# IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
# DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE
# FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
# DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
# SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER
# CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,
# OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
# OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

import clize
import json

import qonnx.analysis.inference_cost as infca
from qonnx.core.datatype import DataType
from qonnx.core.modelwrapper import ModelWrapper
from qonnx.transformation.fold_constants import FoldConstants
from qonnx.transformation.general import (
    GiveReadableTensorNames,
    GiveUniqueNodeNames,
    GiveUniqueParameterTensors,
    RemoveStaticGraphInputs,
    RemoveUnusedTensors,
)
from qonnx.transformation.infer_datatypes import InferDataTypes
from qonnx.transformation.infer_shapes import InferShapes


def compute_bops_and_macs(inf_cost_dict):
    total_bops = 0.0
    total_macs = 0.0
    for k, v in inf_cost_dict.items():
        if k.startswith("op_mac"):
            comps = k.split("_")
            dt1 = DataType[comps[2]]
            dt2 = DataType[comps[3]]
            total_bops += dt1.bitwidth() * dt2.bitwidth() * v
            total_macs += v
    return total_bops, total_macs


def compute_mem_bits_and_elems(inf_cost_dict, filter_string="mem_w"):
    total_mem_bits = 0.0
    total_mem_elems = 0.0
    for k, v in inf_cost_dict.items():
        if k.startswith(filter_string):
            comps = k.split("_")
            dt = DataType[comps[2]]
            total_mem_bits += dt.bitwidth() * v
            total_mem_elems += v
    return total_mem_bits, total_mem_elems


def assign_mem_bits_and_elems(res_dict):
    mem_w_bits, mem_w_elems = compute_mem_bits_and_elems(res_dict, "mem_w")
    mem_o_bits, mem_o_elems = compute_mem_bits_and_elems(res_dict, "mem_o")
    res_dict["total_mem_w_bits"] = mem_w_bits
    res_dict["total_mem_w_elems"] = mem_w_elems
    res_dict["total_mem_o_bits"] = mem_o_bits
    res_dict["total_mem_o_elems"] = mem_o_elems
    return res_dict


def inference_cost(
    model_filename_or_wrapper,
    *,
    output_json=None,
    output_onnx=None,
    preprocess=True,
    discount_sparsity=True,
    cost_breakdown=False
):
    """Return the inference cost estimate metric for given ONNX model.
    Supports the Quant op for weight/activation quantization.

    :param model_filename_or_wrapper: Filename or ModelWrapper for ONNX model
    :param output_json: Optional JSON filename to save the inference cost dict
    :param output_onnx: Optional ONNX filename to save the final model after any
        preprocessing
    :param preprocess: If set, run preprocessing steps such as shape inference,
        datatype inference and constant folding. Strongly recommended.
    :param discount_sparsity: If set, will discount op cost of MAC ops with a
        constant zero weight, and the mem cost of constant zero weights.
    :param cost_breakdown: If set, include per-node (by name) and per-node-type
        breakdowns as part of the returned inference cost dict."""

    combined_results = {}
    if isinstance(model_filename_or_wrapper, ModelWrapper):
        model = model_filename_or_wrapper
    else:
        model = ModelWrapper(model_filename_or_wrapper)
    if preprocess:
        qnt_nodes = model.get_nodes_by_op_type("Quant")
        for qnt_node in qnt_nodes:
            qnt_node.domain = "qonnx.custom_op.general"
        model = model.transform(InferShapes())
        model = model.transform(GiveUniqueParameterTensors())
        model = model.transform(InferDataTypes())
        model = model.transform(FoldConstants(exclude_op_types=[]))
        model = model.transform(RemoveUnusedTensors())
        model = model.transform(RemoveStaticGraphInputs())
        model = model.transform(InferDataTypes())
    model = model.transform(GiveUniqueNodeNames())
    model = model.transform(GiveReadableTensorNames())
    if output_onnx is not None:
        model.save(output_onnx)
    ret = model.analysis(lambda x: infca.inference_cost(x, discount_sparsity, cost_breakdown))
    for i, res in ret.items():
        if i == "total_cost":
            bops, macs = compute_bops_and_macs(res)
            res = assign_mem_bits_and_elems(res)
            res["total_bops"] = bops
            res["total_macs"] = macs
            if "unsupported" in res:
                res["unsupported"] = str(res["unsupported"])
            combined_results[i] = res
        elif i in ["optype_cost", "node_cost"]:
            per_optype_or_node_breakdown = {}
            for optype, op_res in res.items():
                bops, macs = compute_bops_and_macs(op_res)
                op_res = assign_mem_bits_and_elems(op_res)
                op_res["total_bops"] = bops
                op_res["total_macs"] = macs
                per_optype_or_node_breakdown[optype] = op_res
            combined_results[i] = per_optype_or_node_breakdown
    if output_json is not None:
        with open(output_json, "w") as f:
            json.dump(combined_results, f, sort_keys=True, indent=2)
    return combined_results


def main():
    clize.run(inference_cost)


if __name__ == "__main__":
    main()



# Copyright (c) 2020 Xilinx, Inc.
# All rights reserved.
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:
#
# * Redistributions of source code must retain the above copyright notice, this
#   list of conditions and the following disclaimer.
#
# * Redistributions in binary form must reproduce the above copyright notice,
#   this list of conditions and the following disclaimer in the documentation
#   and/or other materials provided with the distribution.
#
# * Neither the name of Xilinx nor the names of its
#   contributors may be used to endorse or promote products derived from
#   this software without specific prior written permission.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
# AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
# IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
# DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE
# FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
# DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
# SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER
# CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,
# OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
# OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

import json
import numpy as np
import warnings

# Protobuf onnx graph node type
from onnx import NodeProto  # noqa
from onnx import mapping
from toposort import toposort_flatten

import qonnx.util.basic as util
from qonnx.transformation.base import Transformation


class MovePadAttributeToTensor(Transformation):
    "Move padding info from attribute into input tensor for Pad nodes."

    def apply(self, model):
        run_again = False
        pad_nodes = model.get_nodes_by_op_type("Pad")
        for pad_node in pad_nodes:
            pads = util.get_by_name(pad_node.attribute, "pads")
            if pads is not None:
                assert len(pad_node.input) == 1
                pads_t = np.asarray(pads.ints)
                new_pad_name = model.make_new_valueinfo_name()
                model.set_initializer(new_pad_name, pads_t)
                pad_node.input.append(new_pad_name)
                pad_node.attribute.remove(pads)
            padval = util.get_by_name(pad_node.attribute, "value")
            if padval is not None:
                # non-float types will need type correction here
                input_vi = model.get_tensor_valueinfo(pad_node.input[0])
                pad_dtype = mapping.TENSOR_TYPE_TO_NP_TYPE[input_vi.type.tensor_type.elem_type]
                padval_t = np.asarray(padval.f, pad_dtype)
                new_padval_name = model.make_new_valueinfo_name()
                model.set_initializer(new_padval_name, padval_t)
                pad_node.input.append(new_padval_name)
                pad_node.attribute.remove(padval)

        return (model, run_again)


class RemoveUnusedTensors(Transformation):
    """Remove any unused tensors in the graph by removing any initializers,
    ValueInfo and tensor annotations associated with it. Unused tensors do not
    appear as any input/output for any graph nodes.
    """

    def apply(self, model):
        graph_modified = False
        onnx_graph = model.model.graph
        # build a set of tensors that we actually use in the graph nodes
        used_tensors = set()
        for node in model.graph.node:
            for i in node.input:
                used_tensors.add(i)
            for o in node.output:
                used_tensors.add(o)
        # remove initializers, value_info and annotations that are not in the
        # used set of tensors, as determined by the graph node i/o
        for init in onnx_graph.initializer:
            if init.name not in used_tensors:
                onnx_graph.initializer.remove(init)
                graph_modified = True
        for vi in onnx_graph.value_info:
            if vi.name not in used_tensors:
                onnx_graph.value_info.remove(vi)
                graph_modified = True
        for qa in onnx_graph.quantization_annotation:
            if qa.tensor_name not in used_tensors:
                onnx_graph.quantization_annotation.remove(qa)
                graph_modified = True

        return (model, graph_modified)


class RemoveStaticGraphInputs(Transformation):
    "Remove any top-level graph inputs that have initializers."

    def apply(self, model):
        graph_modified = False
        for i in model.graph.input:
            if model.get_initializer(i.name) is not None:
                # move ValueInfo to internal (value_info) container
                model.graph.value_info.append(i)
                model.graph.input.remove(i)
                graph_modified = True

        return (model, graph_modified)


class GiveUniqueNodeNames(Transformation):
    """Give unique names to each node in the graph using enumeration, starting
    with given prefix (if specified in the constructor)."""

    def __init__(self, prefix=""):
        super().__init__()
        self.prefix = prefix

    def apply(self, model):
        optype_count = {}
        for n in model.graph.node:
            if n.op_type not in optype_count.keys():
                optype_count[n.op_type] = 0
            n.name = "%s%s_%d" % (self.prefix, n.op_type, optype_count[n.op_type])
            optype_count[n.op_type] += 1
        # return model_was_changed = False as single iteration is always enough
        return (model, False)


class GiveRandomTensorNames(Transformation):
    """Give random tensor names to all tensors."""

    def apply(self, model):
        names = model.get_all_tensor_names()
        for name in names:
            model.rename_tensor(name, util.random_string())
        # return model_was_changed = False as single iteration is always enough
        return (model, False)


class GiveReadableTensorNames(Transformation):
    """Give more human-readable names to all internal tensors. You should
    apply GiveUniqueNodeNames prior to this transform to avoid empty node names,
    as the readable names are based on the node names."""

    def apply(self, model):
        # to ensure we can use rename_tensor safely (without renaming existing
        # tensors) we start by giving random names to all tensors
        model = model.transform(GiveRandomTensorNames())
        graph = model.graph
        for n in graph.node:
            assert n.name != "", "Found empty node name"
            out_num = 0
            for o in n.output:
                model.rename_tensor(o, "%s_out%d" % (n.name, out_num))
                out_num += 1
            init_in_num = 0
            for i in n.input:
                if model.get_initializer(i) is not None:
                    model.rename_tensor(i, "%s_param%d" % (n.name, init_in_num))
                    init_in_num += 1
        # give special names to the model inputs and outputs
        for i, inp in enumerate(model.graph.input):
            iname = "global_in" if i == 0 else "global_in_%d" % i
            model.rename_tensor(inp.name, iname)
        for i, outp in enumerate(model.graph.output):
            oname = "global_out" if i == 0 else "global_out_%d" % i
            model.rename_tensor(outp.name, oname)
        # return model_was_changed = False as single iteration is always enough
        return (model, False)


class GiveUniqueParameterTensors(Transformation):
    """Make every parameter tensor unique. The aim is to avoid affecting
    other nodes apart from the one the system is currently operating on."""

    def apply(self, model):
        graph = model.graph
        graph_modified = False
        seen_parameters = []
        for n in graph.node:
            # copy inputs since they may be modified
            node_inputs_list = [x for x in n.input]
            for input_idx, node_input in enumerate(node_inputs_list):
                # check if it's a parameter
                input_init = model.get_initializer(node_input)
                if input_init is None:
                    # dynamic input
                    continue

                # check if repeated
                if node_input not in seen_parameters:
                    # first occurance
                    seen_parameters += [node_input]
                    continue

                new_param_name = model.make_new_valueinfo_name()

                model.set_initializer(new_param_name, input_init)
                model.set_tensor_datatype(new_param_name, model.get_tensor_datatype(node_input))

                # point node input to new tensor
                n.input[input_idx] = new_param_name

        return (model, graph_modified)


class SortGraph(Transformation):
    """Returns the model with its node list sorted topologically.
    Any ONNX graph to be executed must have a topologically sorted node list,
    as dictated by the ONNX standard.
    """

    # Notes on SortGraph performance:
    # benchmark in  tests/transformation/test_sort_graph.py
    # The algorithm doesn't move initializers so its performance should only depend on
    # the number of nodes
    #
    # Relative order of magnitudes for time per step:
    # - Gather graph structure:       base
    # - Sort nodes:                   0.1 of base
    # - Remove and insert in order :  0.001 of base
    #
    # Notes:
    # Remove nodes and insert them in order:
    # Probably this is faster than copying initializers and more robust in general

    def apply(self, model):
        if len(model.graph.node) == 1:
            # single-node graph, nothing to sort
            return (model, False)
        # Gather graph structure
        graph_dependencies = {}
        node_list = [n for n in model.graph.node]  # I also need the list to remove the nodes
        for node_idx, n in enumerate(node_list):
            node_pred = model.find_direct_predecessors(n)
            if node_pred is None:
                # Will also eliminate nodes that are floating around for some reason
                continue

            node_dependencies = [node_list.index(pred) for pred in node_pred]
            graph_dependencies[node_idx] = set(node_dependencies)

        # Sort nodes
        sorted_node_indexes = toposort_flatten(graph_dependencies)

        # Remove nodes and insert them in order
        # Can't remove nodes before if I want to use model.find_direct_predecessors()
        for n in node_list:
            model.graph.node.remove(n)

        for new_idx, sorted_idx in enumerate(sorted_node_indexes):
            model.graph.node.insert(new_idx, node_list[sorted_idx])

        return (model, False)


class ConvertSubToAdd(Transformation):
    """Convert subtract-a-constant nodes to add-a-constant nodes."""

    def apply(self, model):
        graph = model.graph
        for n in graph.node:
            if n.op_type == "Sub":
                A = model.get_initializer(n.input[1])
                if A is not None:
                    n.op_type = "Add"
                    model.set_initializer(n.input[1], -A)
        # return model_was_changed = False as single iteration is always enough
        return (model, False)


class ConvertDivToMul(Transformation):
    """Convert divide by constant nodes to multiply by constant nodes."""

    def apply(self, model):
        graph = model.graph
        for n in graph.node:
            if n.op_type == "Div":
                A = model.get_initializer(n.input[1])
                if A is not None:
                    n.op_type = "Mul"
                    model.set_initializer(n.input[1], (1.0 / A).astype(A.dtype))
        # return model_was_changed = False as single iteration is always enough
        return (model, False)


class ApplyConfig(Transformation):
    """Applies node properties (attributes) from either a config dict or its JSON
    representation given as a filename.
    The JSON file can specify default values for particular op_types, as well
    as values for nodes with particular names. Example dict::

        {
        # set kernel_size = 3 for all nodes with op_type=Im2Col
        "Defaults" : {"kernel_size" : [3, ["Im2Col"]]},
        # set kernel_size = 7 for the particular node with name Im2Col_0
        "Im2Col_0" : {"kernel_size" : 7}
        }

    """

    def __init__(self, config, node_filter=lambda x: True):
        super().__init__()
        self.config = config
        self.node_filter = node_filter

    def apply(self, model):
        if isinstance(self.config, dict):
            model_config = self.config
        else:
            with open(self.config, "r") as f:
                model_config = json.load(f)

        used_configurations = ["Defaults"]
        missing_configurations = []

        # Configure network
        for node_idx, node in enumerate(model.graph.node):
            if not self.node_filter(node):
                continue
            try:
                node_config = model_config[node.name]
            except KeyError:
                missing_configurations += [node.name]
                node_config = {}

            from qonnx.custom_op.registry import getCustomOp

            try:
                inst = getCustomOp(node)
            except Exception:
                continue
            used_configurations += [node.name]

            # set specified defaults
            default_values = []
            for key, value in model_config["Defaults"].items():
                assert len(value) % 2 == 0
                if key not in model_config:
                    for val, op in zip(value[::2], value[1::2]):
                        default_values.append((key, val, op))
                        assert not (op == "all" and len(value) > 2)
            default_configs = {key: val for key, val, op in default_values if op == "all" or node.op_type in op}
            for attr, value in default_configs.items():
                inst.set_nodeattr(attr, value)

            # set node attributes from specified configuration
            for attr, value in node_config.items():
                inst.set_nodeattr(attr, value)

        # Configuration verification
        if len(missing_configurations) > 0:
            warnings.warn("\nNo HW configuration for nodes: " + ", ".join(missing_configurations))

        unused_configs = [x for x in model_config if x not in used_configurations]
        if len(unused_configs) > 0:
            warnings.warn("\nUnused HW configurations: " + ", ".join(unused_configs))

        # one iteration is enough
        return (model, False)


# Groups inputs by categories, i.e., groups dynamic inputs first, followed by
# initializers. Keeps order of inputs in each category.
def group_inputs_by_category(node: NodeProto, model):  # noqa
    # Select all dynamic inputs, which are those without initializer tensor
    dynamics = [i for i in node.input if model.get_initializer(i) is None]
    # Select all input which are initializers, which, by exclusion, are all
    # those not among the dynamic inputs
    initializers = [i for i in node.input if i not in dynamics]
    # Return lists of dynamic anc initializer inputs
    return dynamics, initializers


# Tidy-Up transformation sorting the inputs to all commutative operations to
# have initializer inputs last
class SortCommutativeInputsInitializerLast(Transformation):
    """
    Sorts inputs of nodes describing commutative operations to have initializer
    inputs last. This order of inputs is assumed by many other transformations.
    """

    # Set of supported commutative operations
    #   TODO: There might be more valid operations
    SUPPORTED_COMMUTATIVE_OPS = {"Add", "Mul", "And", "Or", "Xor", "Sum"}

    # Applies the transform to a whole model graph
    def apply(self, model):  # noqa
        # Get the model graph out of the model wrapper object
        graph = model.graph
        # Keep track of whether the graph has been modified
        graph_modified = False
        # Iterate all nodes in the graph keeping track of the index
        for index, node in enumerate(graph.node):
            # Check whether this node is among the supported
            if node.op_type in self.SUPPORTED_COMMUTATIVE_OPS:
                # Group node inputs by category
                dynamics, initializers = group_inputs_by_category(node, model)
                # Flatten the grouped input list
                inputs = [*dynamics, *initializers]
                # Length of sorted and original input list must match
                assert len(inputs) == len(node.input)
                # Reassigned inputs from sorted categories
                for i, name in enumerate(inputs):
                    # The graph has been modified if any input is reordered
                    if node.input[i] != name:
                        # Note: This is never reset back to False
                        graph_modified = True
                    # Reassign input name at the new index
                    node.input[i] = name
        # Return the transformed model and indicate whether the graph actually
        # has been transformed
        return model, graph_modified



# Copyright (c) 2023 Advanced Micro Devices, Inc.
# All rights reserved.
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:
#
# * Redistributions of source code must retain the above copyright notice, this
#   list of conditions and the following disclaimer.
#
# * Redistributions in binary form must reproduce the above copyright notice,
#   this list of conditions and the following disclaimer in the documentation
#   and/or other materials provided with the distribution.
#
# * Neither the name of qonnx nor the names of its
#   contributors may be used to endorse or promote products derived from
#   this software without specific prior written permission.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
# AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
# IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
# DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE
# FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
# DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
# SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER
# CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,
# OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
# OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

import numpy as np
import onnxruntime as rt
from copy import deepcopy
from onnx import TensorProto, helper
from onnx.helper import make_opsetid

from qonnx.custom_op.base import CustomOp
from qonnx.util.basic import qonnx_make_model


def to_channels_last_args(ndim):
    """
    Returns the tuple of parameters to transpose a channels first tensor to a channels last one.
    :param ndim: Number of dimensions of the tensor to be transposed.
    """
    arg_list = list(range(ndim))
    arg_list.pop(1)
    arg_list.append(1)
    return tuple(arg_list)


def to_channels_first_args(ndim):
    """
    Returns the tuple of parameters to transpose a channels last tensor to a channels first one.
    :param ndim: Number of dimensions of the tensor to be transposed.
    """
    arg_list = list(range(ndim))
    arg_list.pop(-1)
    arg_list.insert(1, ndim - 1)
    return tuple(arg_list)


class ChannelsLastWrappedOp(CustomOp):
    # ToDo: _channelsLast_node_types should be loaded / inferred from this file or the registry.
    # Standard ONNX nodes which require a ChannelsLast data format to function properly
    _channelsLast_node_types = ["Conv", "MaxPool", "BatchNormalization"]

    def infer_node_datatype(self, model):
        # data type stays the same for all supported nodes
        node = self.onnx_node
        dtype = model.get_tensor_datatype(node.input[0])
        model.set_tensor_datatype(node.output[0], dtype)

    def verify_node(self):
        # Check general compatibility
        node = self.onnx_node
        assert (
            node.op_type in self._channelsLast_node_types
        ), f"{node.op_type} is not supported by the ChannelsLast wrapper op."
        assert len(node.input) > 0, "The ChannelsLast wrapper op only supports nodes with inputs."
        assert len(node.output) == 1, "The ChannelsLast wrapper op only supports nodes with exactly one output."

        result = [
            "ONNX OP-type is supported by ChannelsLast wrapper for node execution.",
            "Number of inputs and outputs is valid for node execution.",
        ]

        return result

    def execute_node(self, context, graph):
        node = self.onnx_node

        # Get opset
        opset_version = self.onnx_opset_version
        opset_imports = [make_opsetid("", opset_version)]
        onnx_kwargs = {"opset_imports": opset_imports}

        # Create an intermediate node and remove the domain
        # This enables us to use onnxrutime to execute this node.
        intermediate_node = deepcopy(node)
        intermediate_node.domain = ""

        # Create an intermediate context
        # intermediate_context = {}
        input_dict = {}
        input_tensor_list = []
        output_tensor_list = []

        # Create transposed (channel first) arrays
        # and onnx tensors for the inputs and outputs.
        # And store them in the internal context.
        ndim = len(context[intermediate_node.input[0]].shape)
        for i, input in enumerate(intermediate_node.input):
            channelsFirst_array = context[input]
            # Generally we only transpose the first input
            transpose_input = i < 1
            # Conv is an exception, it also requires the second input to be transposed.
            transpose_input |= intermediate_node.op_type == "Conv" and i < 2
            if transpose_input:
                channelsFirst_array = channelsFirst_array.transpose(to_channels_first_args(ndim))
            assert channelsFirst_array.dtype == np.float32, "Requires float tensor, currently."
            tensor = helper.make_tensor_value_info(input, TensorProto.FLOAT, channelsFirst_array.shape)
            input_dict[input] = channelsFirst_array
            input_tensor_list.append(tensor)

        output = intermediate_node.output[0]
        channelsFirst_array = context[output]
        channelsFirst_array = channelsFirst_array.transpose(to_channels_first_args(ndim))
        assert channelsFirst_array.dtype == np.float32, "Requires float tensor, currently."
        tensor = helper.make_tensor_value_info(output, TensorProto.FLOAT, channelsFirst_array.shape)
        output_tensor_list.append(tensor)

        # Execute the intermediate node with onnxruntime,
        # using the transposed inputs / outputs
        intermediate_graph = helper.make_graph([intermediate_node], "test_model", input_tensor_list, output_tensor_list)
        intermediate_model = qonnx_make_model(intermediate_graph, **onnx_kwargs)
        sess = rt.InferenceSession(intermediate_model.SerializeToString())
        output_list = sess.run(None, input_dict)
        output_onnx = output_list[0]

        # Transpose the output back to channel last and save it in the external context.
        output_onnx = output_onnx.transpose(to_channels_last_args(ndim))
        context[node.output[0]] = output_onnx



# Copyright (c) 2023 Advanced Micro Devices, Inc.
# All rights reserved.
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:
#
# * Redistributions of source code must retain the above copyright notice, this
#   list of conditions and the following disclaimer.
#
# * Redistributions in binary form must reproduce the above copyright notice,
#   this list of conditions and the following disclaimer in the documentation
#   and/or other materials provided with the distribution.
#
# * Neither the name of qonnx nor the names of its
#   contributors may be used to endorse or promote products derived from
#   this software without specific prior written permission.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
# AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
# IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
# DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE
# FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
# DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
# SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER
# CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,
# OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
# OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

import pytest

import numpy as np
import onnx.parser as oprs

from qonnx.core.modelwrapper import ModelWrapper
from qonnx.core.onnx_exec import execute_onnx
from qonnx.transformation.extract_quant_scale_zeropt import ExtractQuantScaleZeroPt


def make_test_model(ishp, channelwise, bitwidth, need_extraction_scale, need_extraction_zeropt):
    ishp_str = str(list(ishp))
    if channelwise:
        q_attr_shp = ishp
    else:
        q_attr_shp = (1,)
    attrshp_str = str(list(q_attr_shp))
    np.random.seed(0)
    if need_extraction_scale:
        scale = np.random.rand(*q_attr_shp).astype(np.float32)
    else:
        scale = np.ones(q_attr_shp, dtype=np.float32)
    if need_extraction_zeropt:
        zeropt = np.random.rand(*q_attr_shp).astype(np.float32)
    else:
        zeropt = np.zeros(q_attr_shp, dtype=np.float32)
    signed = 1
    narrow = 1
    rounding_mode = "ROUND"

    input = f"""
    <
        ir_version: 7,
        opset_import: ["" : 9]
    >
    agraph (float{ishp_str} in0) => (float{ishp_str} out0)
    <
        float{attrshp_str} scale_param,
        float{attrshp_str} zeropt_param,
        float bitwidth_param
    >
    {{
        out0 = qonnx.custom_op.general.Quant<
            signed={str(signed)},
            narrow={str(narrow)},
            rounding_mode="{rounding_mode}"
        >(in0, scale_param, zeropt_param, bitwidth_param)
    }}
    """
    model = oprs.parse_model(input)
    model = ModelWrapper(model)
    model.set_initializer("scale_param", scale)
    model.set_initializer("zeropt_param", zeropt)
    model.set_initializer("bitwidth_param", bitwidth)
    return model


@pytest.mark.parametrize("need_extraction_scale", [True, False])
@pytest.mark.parametrize("need_extraction_zeropt", [True, False])
@pytest.mark.parametrize("channelwise", [True, False])
def test_extract_quant_scale_zeropt(channelwise, need_extraction_scale, need_extraction_zeropt):
    ishp = (1, 10)
    bitwidth = np.asarray(4.0, dtype=np.float32)
    model = make_test_model(ishp, channelwise, bitwidth, need_extraction_scale, need_extraction_zeropt)
    ishp = model.get_tensor_shape("in0")
    inp = np.random.rand(*ishp).astype(np.float32)
    y_golden = execute_onnx(model, {"in0": inp})["out0"]
    model_new = model.transform(ExtractQuantScaleZeroPt())
    y_ret = execute_onnx(model_new, {"in0": inp})["out0"]
    assert np.allclose(y_golden, y_ret)
    qnt_node = model_new.get_nodes_by_op_type("Quant")[0]
    new_scale = model_new.get_initializer(qnt_node.input[1])
    assert (new_scale == 1).all()
    new_zeropt = model_new.get_initializer(qnt_node.input[2])
    assert (new_zeropt == 0).all()
    if need_extraction_scale:
        assert len(model_new.get_nodes_by_op_type("Mul")) == 1
        assert len(model_new.get_nodes_by_op_type("Div")) == 1
    if need_extraction_zeropt:
        assert len(model_new.get_nodes_by_op_type("Add")) == 1
        assert len(model_new.get_nodes_by_op_type("Sub")) == 1



# Copyright (c) 2020, Xilinx
# All rights reserved.
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:
#
# * Redistributions of source code must retain the above copyright notice, this
#   list of conditions and the following disclaimer.
#
# * Redistributions in binary form must reproduce the above copyright notice,
#   this list of conditions and the following disclaimer in the documentation
#   and/or other materials provided with the distribution.
#
# * Neither the name of QONNX nor the names of its
#   contributors may be used to endorse or promote products derived from
#   this software without specific prior written permission.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
# AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
# IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
# DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE
# FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
# DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
# SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER
# CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,
# OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
# OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

from pkgutil import get_data

from qonnx.core.datatype import DataType
from qonnx.core.modelwrapper import ModelWrapper
from qonnx.transformation.fold_constants import FoldConstants
from qonnx.transformation.general import GiveReadableTensorNames, GiveUniqueNodeNames
from qonnx.transformation.infer_datatypes import InferDataTypes, infer_mac_result_dtype
from qonnx.transformation.infer_shapes import InferShapes


def test_infer_mac_dtype_result():
    # dtype prototypes
    is32 = DataType["INT32"]
    iu32 = DataType["UINT32"]
    f32 = DataType["FLOAT32"]
    is4 = DataType["INT4"]
    iu4 = DataType["UINT4"]
    fx4 = DataType["FIXED<4,2>"]
    si4 = DataType["SCALEDINT<4>"]
    si32 = DataType["SCALEDINT<32>"]
    # test several 2-input (e.g. weights, inputs) cases
    assert infer_mac_result_dtype([iu4, iu4], None, False) == iu32
    assert infer_mac_result_dtype([iu4, is4], None, False) == is32
    assert infer_mac_result_dtype([iu4, iu4], None, True) == is32
    assert infer_mac_result_dtype([iu4, fx4], None, False) == si32
    assert infer_mac_result_dtype([fx4, si4], None, False) == si32
    assert infer_mac_result_dtype([is4, si4], None, False) == si32
    assert infer_mac_result_dtype([f32, iu4], f32, False) == f32
    assert infer_mac_result_dtype([f32, si4], f32, False) == f32
    # test several 3-input (e.g. weights, inputs, biases) cases
    assert infer_mac_result_dtype([iu4, iu4, iu4], None, False) == iu32
    assert infer_mac_result_dtype([iu4, iu4, is4], None, False) == is32
    assert infer_mac_result_dtype([is4, iu4, fx4], None, False) == si32
    assert infer_mac_result_dtype([is4, iu4, f32], f32, False) == f32


def test_infer_datatypes():
    raw_m = get_data("qonnx.data", "onnx/mnist-conv/model.onnx")
    model = ModelWrapper(raw_m)
    model = model.transform(InferShapes())
    model = model.transform(FoldConstants())
    model = model.transform(GiveUniqueNodeNames())
    model = model.transform(GiveReadableTensorNames())
    # this model has no DataType info, so add some DataType annotation
    # to make things a bit more exciting
    model.set_tensor_datatype("global_in", DataType["UINT8"])
    # Conv with int weights + inputs will have int output datatype
    model.set_tensor_datatype("Conv_0_param0", DataType["INT4"])
    model = model.transform(InferDataTypes())
    assert model.get_tensor_datatype("global_in") == DataType["UINT8"]
    assert model.get_tensor_datatype("Conv_0_out0") == DataType["INT32"]
    assert model.get_tensor_datatype("Relu_0_out0") == DataType["FLOAT32"]
    assert model.get_tensor_datatype("global_out") == DataType["FLOAT32"]



# Copyright (c) 2020 Xilinx, Inc.
# All rights reserved.
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:
#
# * Redistributions of source code must retain the above copyright notice, this
#   list of conditions and the following disclaimer.
#
# * Redistributions in binary form must reproduce the above copyright notice,
#   this list of conditions and the following disclaimer in the documentation
#   and/or other materials provided with the distribution.
#
# * Neither the name of Xilinx nor the names of its
#   contributors may be used to endorse or promote products derived from
#   this software without specific prior written permission.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
# AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
# IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
# DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE
# FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
# DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
# SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER
# CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,
# OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
# OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

import numpy as np
import onnx.helper as helper

from qonnx.core.datatype import DataType
from qonnx.custom_op.base import CustomOp


def xnorpopcountmatmul(inp0, inp1):
    """Simulates XNOR-popcount matrix multiplication as a regular bipolar
    matrix multiplication followed by some post processing."""
    # extract the operand shapes
    # (M, K0) = inp0.shape
    # (K1, N) = inp1.shape
    K0 = inp0.shape[-1]
    K1 = inp1.shape[0]
    # make sure shapes are compatible with matmul
    assert K0 == K1, "Matrix shapes are not compatible with matmul."
    K = K0
    # convert binary inputs to bipolar
    inp0_bipolar = 2.0 * inp0 - 1.0
    inp1_bipolar = 2.0 * inp1 - 1.0
    # call regular numpy matrix multiplication
    out = np.matmul(inp0_bipolar, inp1_bipolar)
    # XNOR-popcount does not produce the regular dot product result --
    # it returns the number of +1s after XNOR. let P be the number of +1s
    # and N be the number of -1s. XNOR-popcount returns P, whereas the
    # regular dot product result from numpy is P-N, so we need to apply
    # some correction.
    # out = P-N
    # K = P+N
    # out + K = 2P, so P = (out + K)/2
    return (out + K) * 0.5


class XnorPopcountMatMul(CustomOp):
    """Class that corresponds to a XNOR-popcount matrix
    multiplication node."""

    def get_nodeattr_types(self):
        return {}

    def make_shape_compatible_op(self, model):
        node = self.onnx_node
        return helper.make_node("MatMul", [node.input[0], node.input[1]], [node.output[0]])

    def infer_node_datatype(self, model):
        node = self.onnx_node
        # ensure inputs are binary
        assert (
            model.get_tensor_datatype(node.input[0]) == DataType["BINARY"]
        ), """QONNX
        DataType of first input is not set to BINARY as it should be."""
        assert (
            model.get_tensor_datatype(node.input[1]) == DataType["BINARY"]
        ), """QONNX
        DataTypes of second input is not set to BINARY as it should be."""
        # XNOR-popcount produces unsigned integers, assume uint32
        model.set_tensor_datatype(node.output[0], DataType["UINT32"])

    def execute_node(self, context, graph):
        node = self.onnx_node
        # save inputs
        inp0 = context[node.input[0]]
        inp1 = context[node.input[1]]
        # calculate output
        output = xnorpopcountmatmul(inp0, inp1)
        # set context according to output name
        context[node.output[0]] = output

    def verify_node(self):
        info_messages = []

        # verify number of attributes
        num_of_attr = 0
        if len(self.onnx_node.attribute) == num_of_attr:
            info_messages.append("The number of attributes is correct")
        else:
            info_messages.append(
                """The number of attributes is incorrect,
            {} should have {} attributes""".format(
                    self.onnx_node.op_type, num_of_attr
                )
            )

        # verify that all necessary attributes exist
        info_messages.append("XnorPopcountMatMul should not have any attributes")

        # verify the number of inputs
        if len(self.onnx_node.input) == 2:
            info_messages.append("The number of inputs is correct")
        else:
            info_messages.append("XnorPopcountMatMul needs 2 data inputs")

        return info_messages



# Copyright (c) 2021, Xilinx
# All rights reserved.
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:
#
# * Redistributions of source code must retain the above copyright notice, this
#   list of conditions and the following disclaimer.
#
# * Redistributions in binary form must reproduce the above copyright notice,
#   this list of conditions and the following disclaimer in the documentation
#   and/or other materials provided with the distribution.
#
# * Neither the name of QONNX nor the names of its
#   contributors may be used to endorse or promote products derived from
#   this software without specific prior written permission.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
# AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
# IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
# DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE
# FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
# DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
# SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER
# CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,
# OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
# OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

import warnings
from onnx import helper

from qonnx.transformation.base import Transformation


class ExtractBiasFromConv(Transformation):
    """
    Extracts the (optional) Bias from a Conv(Transpose) node and inserts it behind the
    Conv(Transpose) node as an Add node.
    """

    def apply(self, model):
        graph = model.graph
        node_ind = 0
        for n in graph.node:
            node_ind += 1
            if n.op_type in ["Conv", "ConvTranspose"]:
                # Check if the node has a bias input
                if len(n.input) > 2:
                    # Extract bias
                    bias = model.get_initializer(n.input[2])
                    if bias is None:
                        warnings.warn(f"Could not extract bias from node {n}")
                        continue

                    # Insert bias as Add node behind the Conv node
                    out_shape = model.get_tensor_shape(n.output[0])
                    # Reshape bias tensor
                    add_shape = [1] * len(out_shape)
                    # ToDo: this must change to "add_shape[-1] = bias.shape[0]" when
                    #  the channels last layout comes around.
                    bias_shape = model.get_tensor_shape(n.input[2])
                    add_shape[1] = bias_shape[0]
                    if bias is not None:
                        model.set_initializer(n.input[2], bias.reshape(add_shape))

                    act_add_tensor = helper.make_tensor_value_info(
                        model.make_new_valueinfo_name(),
                        model.get_tensor_valueinfo(n.output[0]).type.tensor_type.elem_type,
                        out_shape,
                    )
                    graph.value_info.append(act_add_tensor)

                    add_node = helper.make_node(
                        "Add",
                        [act_add_tensor.name, n.input[2]],
                        [n.output[0]],
                    )
                    graph.node.insert(node_ind, add_node)

                    # Repoint Conv output and remove bias tensor
                    n.output[0] = act_add_tensor.name
                    n.input.remove(n.input[2])

                    return model, True

        return model, False



# Copyright (c) 2020, Xilinx
# All rights reserved.
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:
#
# * Redistributions of source code must retain the above copyright notice, this
#   list of conditions and the following disclaimer.
#
# * Redistributions in binary form must reproduce the above copyright notice,
#   this list of conditions and the following disclaimer in the documentation
#   and/or other materials provided with the distribution.
#
# * Neither the name of QONNX nor the names of its
#   contributors may be used to endorse or promote products derived from
#   this software without specific prior written permission.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
# AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
# IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
# DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE
# FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
# DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
# SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER
# CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,
# OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
# OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

import numpy as np
import onnx
import onnx.numpy_helper as np_helper
from pkgutil import get_data

import qonnx.core.onnx_exec as oxe
from qonnx.core.modelwrapper import ModelWrapper
from qonnx.transformation.fold_constants import FoldConstants
from qonnx.transformation.infer_shapes import InferShapes


def test_const_folding():
    raw_m = get_data("qonnx.data", "onnx/mnist-conv/model.onnx")
    model = ModelWrapper(raw_m)
    model = model.transform(InferShapes())
    model = model.transform(FoldConstants())
    raw_i = get_data("qonnx.data", "onnx/mnist-conv/test_data_set_0/input_0.pb")
    raw_o = get_data("qonnx.data", "onnx/mnist-conv/test_data_set_0/output_0.pb")
    input_tensor = onnx.load_tensor_from_string(raw_i)
    output_tensor = onnx.load_tensor_from_string(raw_o)
    input_dict = {"Input3": np_helper.to_array(input_tensor)}
    output_dict = oxe.execute_onnx(model, input_dict)
    assert np.isclose(np_helper.to_array(output_tensor), output_dict["Plus214_Output_0"], atol=1e-3).all()


def test_const_folding_shapes():
    raw_m = get_data("qonnx.data", "onnx/mnist-conv/model.onnx")
    model = ModelWrapper(raw_m)
    model = model.transform(InferShapes())
    mm_node_w_in = model.get_nodes_by_op_type("MatMul")[0].input[1]
    assert model.find_producer(mm_node_w_in) is not None
    assert model.find_producer(mm_node_w_in).op_type == "Reshape"
    assert model.get_initializer(mm_node_w_in) is None
    model = model.transform(FoldConstants())
    assert model.find_producer(mm_node_w_in) is None
    assert model.get_initializer(mm_node_w_in) is not None



import qkeras
import six


def get_quant_params(tensor, qkeras_quantizer):
    if isinstance(qkeras_quantizer, str):
        qkeras_quantizer = qkeras.get_quantizer(qkeras_quantizer)

    return handler_map[qkeras_quantizer.__class__.__name__](tensor, qkeras_quantizer)


def _get_scale_from_alpha(tensor, quantizer):
    alpha = quantizer.get_config()["alpha"]

    if alpha is None:
        return 1
    elif isinstance(alpha, six.string_types):
        raise Exception(f"Cannot parse alpha = {alpha}.")
        return 1
    else:
        return alpha


def _get_quantizer_scale(tensor, quantizer):
    # call the quantizer on the tensor to get its scale
    import numpy as np

    quantizer(np.array(tensor).astype(np.float32))
    return quantizer.scale


def convert_quantized_bits(tensor, quantizer):
    config = quantizer.get_config()
    signed = int(config["keep_negative"])
    narrow = int(config["symmetric"])
    qscale = _get_quantizer_scale(tensor, quantizer)
    assert qscale == 1, "Non-unity alpha is not yet supported"
    scale = 1.0 / 2 ** (int(config["bits"]) - int(config["integer"] + signed))
    zero_point = 0
    bit_width = int(config["bits"])
    rounding_mode = "ROUND"

    settings = {
        "attributes": {"signed": signed, "narrow": narrow, "rounding_mode": rounding_mode},
        "inputs": {"scale": scale, "zero_point": zero_point, "bit_width": bit_width},
    }
    return settings


def convert_quantized_relu(tensor, quantizer):
    config = quantizer.get_config()

    signed = int(config["negative_slope"] != 0.0)
    narrow = int(False)
    scale = 1.0 / 2 ** (int(config["bits"]) - int(config["integer"] + signed))
    zero_point = 0
    bit_width = int(config["bits"])
    rounding_mode = "ROUND"

    settings = {
        "attributes": {"signed": signed, "narrow": narrow, "rounding_mode": rounding_mode},
        "inputs": {"scale": scale, "zero_point": zero_point, "bit_width": bit_width},
    }
    return settings


def convert_binary(tensor, quantizer):
    signed = 1
    narrow = 1
    qscale = _get_quantizer_scale(tensor, quantizer)
    assert qscale == 1, "binary - non-unity alpha is not yet supported"
    scale = 1
    zero_point = 0
    bit_width = 1
    rounding_mode = "ROUND"

    settings = {
        "attributes": {"signed": signed, "narrow": narrow, "rounding_mode": rounding_mode},
        "inputs": {"scale": scale, "zero_point": zero_point, "bit_width": bit_width},
    }
    return settings


def convert_ternary(tensor, quantizer):
    config = quantizer.get_config()
    signed = 1
    narrow = 1
    qscale = _get_quantizer_scale(tensor, quantizer)
    assert qscale == 1, "ternary - non-unity alpha is not yet supported"
    # qkeras ternary quantizer has threshold parameter to change rounding point
    # here we could scale such that normal 'ROUND' op gives the same result, but doesn't work with re-scaling
    t = config["threshold"]
    if t is None:
        ternary = qkeras.ternary()
        t = ternary.default_threshold
    assert t == 0.5, "ternary - only threshold 0.5 is supported"
    # note that if assertions fail, Quant node is not inserted, but model is still converted
    # this seems to be unexpected behavior
    scale = 1.0
    zero_point = 0
    bit_width = 2
    rounding_mode = "ROUND"

    settings = {
        "attributes": {"signed": signed, "narrow": narrow, "rounding_mode": rounding_mode},
        "inputs": {"scale": scale, "zero_point": zero_point, "bit_width": bit_width},
    }
    return settings


handler_map = {
    "quantized_bits": convert_quantized_bits,
    "quantized_relu": convert_quantized_relu,
    "binary": convert_binary,
    "ternary": convert_ternary,
}



# Copyright (c) 2023 Advanced Micro Devices, Inc.
# All rights reserved.
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:
#
# * Redistributions of source code must retain the above copyright notice, this
#   list of conditions and the following disclaimer.
#
# * Redistributions in binary form must reproduce the above copyright notice,
#   this list of conditions and the following disclaimer in the documentation
#   and/or other materials provided with the distribution.
#
# * Neither the name of qonnx nor the names of its
#   contributors may be used to endorse or promote products derived from
#   this software without specific prior written permission.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
# AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
# IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
# DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE
# FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
# DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
# SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER
# CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,
# OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
# OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

import numpy as np
import warnings
from copy import deepcopy
from onnx import TensorProto, helper

from qonnx.core.modelwrapper import ModelWrapper
from qonnx.custom_op import channels_last
from qonnx.custom_op.channels_last.base_wrapped_op import to_channels_first_args, to_channels_last_args
from qonnx.transformation.base import Transformation
from qonnx.transformation.fold_constants import FoldConstants
from qonnx.transformation.general import SortGraph
from qonnx.transformation.infer_shapes import InferShapes
from qonnx.transformation.make_input_chanlast import MakeInputChannelsLast
from qonnx.transformation.quant_constant_folding import FoldTransposeIntoQuantInit
from qonnx.util.basic import get_by_name
from qonnx.util.onnx import is_eltwise_optype

# Standard ONNX nodes which require a ChannelsLast data format to function properly
_channelsLast_node_types = list(channels_last.custom_op.keys())

# Nodes, which do not modify the shape of the tensor
# And modify all values in the same way.
_move_through_nodes = ["Quant", "Relu", "Selu", "LeakyRelu", "Sigmoid", "Tanh"]

# Nodes, which do not modify the shape of the tensor,
# And modify all values in the same way, if the second tensor is a scalar.
_move_through_nodes_if_scalar = ["Mul", "Div", "Sub", "Add"]


def get_transpose_perms(transpose_node, model):
    perm = get_by_name(transpose_node.attribute, "perm")
    ndim = len(model.get_tensor_shape(transpose_node.input[0]))
    if perm is None:
        # default perm is to reverse the dim order
        return list(range(ndim - 1, -1, -1))
    else:
        return list(perm.ints)


def move_transpose_past_eltwise(transpose_node, eltwise_node, model: ModelWrapper):
    t0 = transpose_node.input[0]
    t1 = transpose_node.output[0]
    t2 = eltwise_node.output[0]
    subgraph_inp_shape = model.get_tensor_shape(t0)
    ndim_inp = len(subgraph_inp_shape)
    perm = get_transpose_perms(transpose_node, model)

    # before: t0 -> transpose -> t1 -> eltwise -> t2
    # after: t0 -> eltwise -> t1 -> transpose -> t2
    # find the eltwise inp index fed by transpose
    transpose_in_ind = list(eltwise_node.input).index(t1)
    # check all inputs for the eltwise op:
    # we need to ensure those inputs get inverse-transposed
    # to keep the graph semantics intact
    for ind, eltwise_inp in enumerate(eltwise_node.input):
        if ind == transpose_in_ind:
            # the input that feeds from the original transpose
            # node will be implicitly inverse-transposed, since we'll be
            # moving that transpose node past the eltwise op
            continue
        inp_shape = model.get_tensor_shape(eltwise_inp)
        ndim = len(inp_shape)
        if ndim == 0:
            # scalar input, always broadcastable, no action needed
            continue
        elif ndim == ndim_inp:
            # input with matching dimensions, add inverse transpose
            new_t_inp = model.make_new_valueinfo_name()
            inv_perm = np.argsort(perm)
            new_transpose_node = helper.make_node("Transpose", [eltwise_inp], [new_t_inp], perm=inv_perm)
            t_shape = np.transpose(np.empty(inp_shape), axes=inv_perm).shape
            model.set_tensor_shape(new_t_inp, t_shape)
            eltwise_node.input[ind] = new_t_inp
            model.graph.node.append(new_transpose_node)
        else:
            # input with non-matching dimensions, assume broadcastable
            # first add Unsqueeze node to match number of dimensions
            unsqueeze_param_name = model.make_new_valueinfo_name()
            model.set_initializer(unsqueeze_param_name, np.asarray(list(range(ndim_inp - ndim)), dtype=np.int64))
            unsqueeze_out_name = model.make_new_valueinfo_name()
            new_unsqueeze_node = helper.make_node("Unsqueeze", [eltwise_inp, unsqueeze_param_name], [unsqueeze_out_name])
            unsqueeze_out_shape = np.expand_dims(np.empty(inp_shape), axis=tuple(range(ndim_inp - ndim))).shape
            model.set_tensor_shape(unsqueeze_out_name, unsqueeze_out_shape)
            model.graph.node.append(new_unsqueeze_node)
            # now add inverse transpose
            new_t_inp = model.make_new_valueinfo_name()
            inv_perm = np.argsort(perm)
            new_transpose_node = helper.make_node("Transpose", [unsqueeze_out_name], [new_t_inp], perm=inv_perm)
            t_shape = np.transpose(np.empty(unsqueeze_out_shape), axes=inv_perm).shape
            model.set_tensor_shape(new_t_inp, t_shape)
            eltwise_node.input[ind] = new_t_inp
            model.graph.node.append(new_transpose_node)
    # rewire to swap transpose and eltwise node order
    eltwise_node.input[transpose_in_ind] = t0
    eltwise_node.output[0] = t1
    transpose_node.input[0] = t1
    transpose_node.output[0] = t2
    # t1 tensor shape changes to inp_shape
    model.set_tensor_shape(t1, subgraph_inp_shape)
    model = model.transform(SortGraph())
    model = model.transform(FoldConstants())
    return model


class ConvertToChannelsLastAndClean(Transformation):
    """
    Converts data layout dependent nodes to ChannelsLast nodes and inserts transformations.
    Then it tries to eliminate as many transformations as possible and moves the
    still existing ones as far upstream as possible.

    :param make_input_channels_last: Also makes the input of the network channels last,
        otherwise a transpose node will be left at the beginning of the network. Defaults to False
    :type make_input_channels_last: bool

    """

    def __init__(self, make_input_channels_last=False):
        super().__init__()
        self._make_input_channels_last = make_input_channels_last

    def apply(self, model: ModelWrapper):
        assert model.check_all_tensor_shapes_specified(), (
            "All tensor shapes must be specified. " "Consider running InferShapes."
        )
        model = model.transform(InsertChannelsLastDomainsAndTrafos())
        initial_model_string = model.model.SerializeToString()
        # Apply RemoveConsecutiveChanFirstAndChanLastTrafos
        model = model.transform(RemoveConsecutiveChanFirstAndChanLastTrafos())

        # Apply MoveChanLastUpstream and fold into initializers
        model = model.transform(MoveChanLastUpstream())
        model = model.transform(FoldTransposeIntoQuantInit())

        # Run RemoveConsecutiveChanFirstAndChanLastTrafos again,
        # Technically only required if something changed in the previous trafo
        model = model.transform(RemoveConsecutiveChanFirstAndChanLastTrafos())

        # Apply MoveChanLastDownStream and MoveTransposePastFork
        model = model.transform(MoveChanFirstDownstream())
        model = model.transform(MoveTransposePastFork())

        # Run RemoveConsecutiveChanFirstAndChanLastTrafos again,
        # Technically only required if something changed in the previous trafo
        model = model.transform(RemoveConsecutiveChanFirstAndChanLastTrafos())

        # Apply AbsorbChanFirstIntoMatMul
        model = model.transform(AbsorbChanFirstIntoMatMul())

        if self._make_input_channels_last:
            model = model.transform(MakeInputChannelsLast())
            model = model.transform(RemoveConsecutiveChanFirstAndChanLastTrafos())

        # Check if the model changed
        new_model_string = model.model.SerializeToString()

        # Do small cleanup, which isn't done by the cleanup in the normal transformation
        model = model.transform(InferShapes())
        model = model.transform(FoldConstants())

        # Check if the model changed
        model_changed = initial_model_string != new_model_string

        return model, model_changed


class InsertChannelsLastDomainsAndTrafos(Transformation):
    """
    Inserts ChannelsLast domain, where required and also inserts required transposes.
    """

    def apply(self, model):
        graph = model.graph
        node_ind = 0
        graph_modified = False
        # Find nodes, where the domain should be changed
        for n in graph.node:
            node_ind += 1
            if (n.op_type in _channelsLast_node_types) and (n.domain == ""):
                running_node_index = node_ind
                # Insert transformation nodes for input nodes
                input_tensors = n.input
                # Skip for BatchNorm and 2D input tensors,
                # these contain only channels and need no transpose.
                chanFirst_shape = model.get_tensor_shape(input_tensors[0])
                if n.op_type == "BatchNormalization" and len(chanFirst_shape) == 2:
                    continue

                for i, inp in enumerate(input_tensors):
                    # Skip higher "order" inputs of the Batch-Norm,
                    # these don't need a transpose.
                    if n.op_type == "BatchNormalization" and i > 0:
                        continue
                    # Skip Conv bias since it doesn't need a transpose
                    if n.op_type == "Conv" and i == 2:
                        continue
                    # Get the shape of the input tensor
                    # and convert it to the shape for the intermediate tensor
                    chanFirst_shape = model.get_tensor_shape(inp)
                    ndim = len(chanFirst_shape)
                    assert ndim == 3 or ndim == 4, "Channels last conversion is only available for 3D and 4D tensors."
                    chanLast_shape = [chanFirst_shape[idx] for idx in to_channels_last_args(ndim)]
                    # Intermediate tensor
                    inp_trans_out = helper.make_tensor_value_info(
                        model.make_new_valueinfo_name(),
                        TensorProto.FLOAT,
                        chanLast_shape,
                    )
                    graph.value_info.append(inp_trans_out)
                    inp_trans_out = inp_trans_out.name

                    # channels last transpose
                    inp_trans_node = helper.make_node("Transpose", [inp], [inp_trans_out], perm=to_channels_last_args(ndim))
                    graph.node.insert(running_node_index, inp_trans_node)
                    running_node_index += 1

                    # Attach to original node
                    n.input[i] = inp_trans_out

                # Insert transformation nodes for output nodes
                output_tensors = n.output
                for i, outp in enumerate(output_tensors):
                    chanFirst_shape = model.get_tensor_shape(outp)
                    ndim = len(chanFirst_shape)
                    assert ndim == 3 or ndim == 4, "Channels last conversion is only available for 3D and 4D tensors."
                    chanLast_shape = [chanFirst_shape[idx] for idx in to_channels_last_args(ndim)]
                    # Intermediat tensor
                    outp_trans_in = helper.make_tensor_value_info(
                        model.make_new_valueinfo_name(),
                        TensorProto.FLOAT,
                        chanLast_shape,
                    )
                    graph.value_info.append(outp_trans_in)
                    outp_trans_in = outp_trans_in.name

                    # ChannelsFirst -> ChannelsLast transpose
                    outp_trans_node = helper.make_node(
                        "Transpose", [outp_trans_in], [outp], perm=to_channels_first_args(ndim)
                    )
                    graph.node.insert(running_node_index, outp_trans_node)
                    running_node_index += 1

                    # Attach to original node
                    n.output[i] = outp_trans_in

                # Modify domain
                n.domain = "qonnx.custom_op.channels_last"
                # Set modified flag
                graph_modified = True

        return model, graph_modified


class RemoveConsecutiveChanFirstAndChanLastTrafos(Transformation):
    """
    Remove two consecutive transformations, which would do:
    (ChannelsLast -> ChannelsFirst) -> (ChannelsFirst -> ChannelsLast)
    Or more concrete, the first converts to channels first and the second to channels last.
    """

    def apply(self, model):
        graph = model.graph
        node_ind = 0
        graph_modified = False

        # Find fitting transpose node pairs
        for n in graph.node:
            node_ind += 1
            if n.op_type == "Transpose":
                # Check the input shape and make sure we support it
                input_shape = model.get_tensor_shape(n.input[0])
                # Check that this is a "to chan first" trafo
                perm_1 = get_transpose_perms(n, model)
                ndim = len(input_shape)
                if list(to_channels_first_args(ndim)) == perm_1:
                    successor_nodes = model.find_direct_successors(n)
                    # skip if:
                    # - this Transpose has no successors (nothing to do)
                    # - this Transpose output is forking (cannot remove)
                    if successor_nodes is None or len(successor_nodes) > 1:
                        continue
                    successor_node = successor_nodes[0]

                    if successor_node.op_type == "Transpose":
                        # Check that this is a "to chan last" trafo,
                        # if so both can get removed.
                        perm_2 = get_transpose_perms(successor_node, model)
                        if list(to_channels_last_args(ndim)) == perm_2:
                            # Connect original input to new output
                            input_tensor = n.input[0]
                            output_tensor_name = successor_node.output[0]

                            target_nodes = model.find_direct_successors(successor_node)
                            if target_nodes is None:
                                continue

                            target_node = target_nodes[0]
                            for i, inp in enumerate(target_node.input):
                                if inp == output_tensor_name:
                                    target_node.input[i] = input_tensor

                            # remove old nodes
                            graph.node.remove(n)
                            graph.node.remove(successor_node)

                            graph_modified = True
        return model, graph_modified


class MoveChanLastUpstream(Transformation):
    """
    Moves channel last transformations further upstream.
    """

    def apply(self, model: ModelWrapper):
        graph = model.graph
        node_ind = 0
        graph_modified = False
        # Find transpose nodes, which are "to chan last" trafos
        for n in graph.node:
            node_ind += 1
            if n.op_type == "Transpose":
                # Check the input shape and make sure we support it
                input_shape = model.get_tensor_shape(n.input[0])
                ndim = len(input_shape)
                perm = get_transpose_perms(n, model)
                if list(to_channels_last_args(ndim)) == perm:
                    predecessors = model.find_direct_predecessors(n)
                    # Check if we reached the top of the graph
                    if predecessors is None:
                        continue
                    predecessor = predecessors[0]

                    # Check if we can simply move through the previous node
                    move_through_valid = predecessor.op_type in _move_through_nodes
                    # Check if we have a node, which applies a scalar change,
                    # then we can also move through.
                    if predecessor.op_type in _move_through_nodes_if_scalar:
                        second_inp_shape = model.get_tensor_shape(predecessor.input[1])
                        if second_inp_shape == [1] or second_inp_shape == []:
                            move_through_valid |= True

                    # don't move through if the predecessor output is a fork
                    if model.is_fork_node(predecessor):
                        move_through_valid = False

                    # Apply move through trafo if possible
                    if move_through_valid:
                        # Input tensors are always input 0
                        inp = predecessor.input[0]
                        if isinstance(model.get_initializer(inp), type(None)):
                            # Swap around node "predecessor" and "n"
                            # collect tensors
                            tensor_1 = inp
                            tensor_2 = n.input[0]
                            tensor_3 = n.output[0]
                            # Now connect the tensors to the nodes again,
                            # but in different order
                            n.input[0] = tensor_1
                            n.output[0] = tensor_2
                            predecessor.input[0] = tensor_2
                            predecessor.output[0] = tensor_3

                            # Change the shape of the middle tensor
                            target_shape = model.get_tensor_shape(tensor_3)
                            model.set_tensor_shape(tensor_2, target_shape)

                            graph_modified = True
                            return model, graph_modified
                        # else:
                        #     # Explicitly apply the transpose to the initializer
                        #     # of the previous node
                        #     target_tensor = model.get_initializer(inp)
                        #     target_tensor = target_tensor.transpose(perm.ints)
                        #     model.set_initializer(inp, target_tensor)
                        #     # Reconnect predecessor and delete transpose node
                        #     predecessor.output[0] = n.output[0]
                        #     graph.node.remove(n)
                        #
                        #     graph_modified = True
                        # return model, graph_modified

        return model, graph_modified


class MoveChanFirstDownstream(Transformation):
    """
    Moves channel first transformations further downstream.
    """

    def apply(self, model):
        graph = model.graph
        node_ind = 0
        graph_modified = False
        # Find transpose nodes, which are "to chan first" trafos
        for node in graph.node:
            node_ind += 1
            if node.op_type == "Transpose":
                # Check the input shape and make sure we support it
                input_shape = model.get_tensor_shape(node.input[0])
                ndim = len(input_shape)
                perm = get_transpose_perms(node, model)
                if list(to_channels_first_args(ndim)) == perm:
                    # Do not move the node, if it is at the top of the graph,
                    # this is a strange edge case, for 1D networks, where channels last and channels first trafos
                    # are identical.
                    predecessors = model.find_direct_predecessors(node)
                    if predecessors is None:
                        continue

                    successors = model.find_direct_successors(node)
                    if successors is None:
                        continue
                    successor = successors[0]
                    transpose_node = node

                    if is_eltwise_optype(successor.op_type):
                        model = move_transpose_past_eltwise(transpose_node, successor, model)
                        graph_modified = True
                        return model, graph_modified

        return model, graph_modified


class AbsorbChanFirstIntoMatMul(Transformation):
    """
    Removes a transpose to channels first node if it is in front of a Flatten and
    MatMul (or Gemm) node.

    The channels first transpose is fused into the initializer of the Quant node acting
    as a weight tensor for the MatMul/Gemm node.
    Reshape nodes with shape [1, -1] are also supported instead of Flatten nodes.
    Independent of whether the flattening operation was performed by a Flatten node
    or a Resphape node, a Flatten node will be reinserted in-front of the MatMul node.

    Note: This transformation removes some of the tensor shapes on the down-stream path.
     Thus running shape inference afterwards is advised.
    """

    def apply(self, model):
        graph = model.graph
        graph_modified = False
        node_ind = 0
        for n in graph.node:
            node_ind += 1
            # also support implicit flatten via reshape, e.g. reshape(1,-1)
            if n.op_type == "Flatten" or n.op_type == "Reshape":
                ishape = model.get_tensor_shape(n.input[0])
                oshape = model.get_tensor_shape(n.output[0])
                if ishape is None or oshape is None:
                    continue
                if len(oshape) == 2 and ishape[0] == oshape[0]:
                    producer = model.find_producer(n.input[0])
                    if (producer is not None) and (producer.op_type == "Transpose"):
                        # transpose + flatten, absorb into following node
                        transp_node = producer
                        # Check the input shape and make sure we support it
                        input_shape = model.get_tensor_shape(transp_node.input[0])
                        # check if transpose converts ChannelsLast to ChannelsFirst
                        ndim = len(input_shape)
                        perms = get_transpose_perms(transp_node, model)
                        if list(to_channels_first_args(ndim)) == perms:
                            producer = model.find_producer(transp_node.input[0])
                            consumer = model.find_consumer(n.output[0])
                            if (consumer is not None) and ((consumer.op_type == "MatMul") or (consumer.op_type == "Gemm")):
                                # Gemm supports transposes on B, so we need to account for that
                                b_transposed = False
                                if consumer.op_type == "Gemm":
                                    b_transposed = bool(get_by_name(consumer.attribute, "transB"))

                                b_shape = model.get_tensor_shape(consumer.input[1])
                                if b_transposed:
                                    b_shape = list(reversed(b_shape))
                                mw = b_shape[0]
                                mh = b_shape[1]
                                if ndim == 4:
                                    (b, h, w, c) = model.get_tensor_shape(transp_node.input[0])
                                elif ndim == 3:
                                    (b, h, c) = model.get_tensor_shape(transp_node.input[0])
                                    w = 1
                                else:
                                    raise ValueError(
                                        f"Inputs of dimensionality ndim={ndim}, are currently not supported "
                                        f"for merging transposes into the matrix multiply weight tensor"
                                    )
                                # Get the weight initilizer
                                quant_node = model.find_producer(consumer.input[1])
                                if quant_node is None:
                                    W = model.get_initializer(consumer.input[1])
                                else:
                                    if quant_node.op_type == "Quant":
                                        W = model.get_initializer(quant_node.input[0])
                                    else:
                                        warnings.warn(
                                            f"Could not find weight initializer for " f"MatMul/Gemm: {consumer.name}"
                                        )
                                        continue
                                if b_transposed:
                                    W = W.T
                                W_new = W.reshape(c, h, w, mh)
                                W_new = W_new.transpose((1, 2, 0, 3))
                                W_new = W_new.reshape(mw, mh)
                                if b_transposed:
                                    W_new = W_new.T
                                if quant_node is None:
                                    model.set_initializer(consumer.input[1], W_new)
                                else:
                                    model.set_initializer(quant_node.input[0], W_new)
                                # remove transpose & flatten nodes
                                consumer.input[0] = transp_node.input[0]
                                graph.node.remove(n)
                                graph.node.remove(transp_node)

                                # Insert a Flatten node in front of the MatMul
                                inp_tensor_name = consumer.input[0]
                                # Intermediate tensor
                                out_tensor = helper.make_tensor_value_info(
                                    model.make_new_valueinfo_name(),
                                    TensorProto.FLOAT,
                                    None,
                                )
                                graph.value_info.append(out_tensor)
                                out_tensor_name = out_tensor.name

                                # Attach to MatMul output
                                consumer.input[0] = out_tensor_name

                                # Flatten node
                                flat_node = helper.make_node(
                                    "Flatten",
                                    [inp_tensor_name],
                                    [out_tensor_name],
                                    axis=1,
                                )
                                graph.node.insert(node_ind, flat_node)

                                graph_modified = True
                            else:
                                warnings.warn(
                                    "Could not absorb transpose->flatten \
                                    into subsequent node"
                                )
        return model, graph_modified


class MoveOpPastFork(Transformation):
    """Move node operations past graph forks. Used when a node before a fork
    can be merged with nodes in the branches
    """

    def __init__(self, op_name_list):
        super().__init__()
        self.ops_to_move = op_name_list

    def apply(self, model):
        graph = model.graph
        graph_modified = False
        nodes = [n for n in graph.node]
        node_ind = 0
        for node in nodes:
            node_ind += 1
            if node.op_type in self.ops_to_move and model.is_fork_node(node) and not model.is_join_node(node):
                # Restrict this transform to operations with constant parameters
                # Assuming parameters is in input 1
                if len(node.input) > 1:
                    op_init_param = model.get_initializer(node.input[1])
                else:
                    op_init_param = None

                # Check case when branches are empty and go
                # to the same node
                consumers = model.find_consumers(node.output[0])
                assert len(consumers) > 1, "Must have >1 consumer"
                unique_consumer = True
                for consum_node in consumers[1:]:
                    if consumers[0] != consum_node:
                        unique_consumer = False
                        break

                if unique_consumer:
                    continue

                for consumer_node in consumers[1:]:
                    # create new node
                    new_output_tensor_name = model.make_new_valueinfo_name()
                    if op_init_param is None:
                        new_inp_list = [node.input[0]]
                    else:
                        new_param_name = model.make_new_valueinfo_name()
                        new_inp_list = [node.input[0], new_param_name]
                        model.set_initializer(new_param_name, op_init_param)
                    new_node = deepcopy(node)
                    new_node.input[:] = new_inp_list
                    new_node.output[:] = [new_output_tensor_name]
                    graph.node.insert(node_ind, new_node)
                    node_ind += 1

                    # change consumer input tensor
                    graph.node.remove(consumer_node)
                    for idx, consumer_input in enumerate(consumer_node.input):
                        if consumer_input == node.output[0]:
                            consumer_node.input[idx] = new_output_tensor_name
                            break
                    else:
                        raise Exception("Consumer should have the current node output as input")

                    graph.node.insert(node_ind, consumer_node)

                graph_modified = True

        model = model.transform(InferShapes())
        return (model, graph_modified)


class MoveAddPastFork(MoveOpPastFork):
    def __init__(self):
        super().__init__(["Add"])


class MoveMulPastFork(MoveOpPastFork):
    def __init__(self):
        super().__init__(["Mul"])


class MoveLinearPastFork(MoveOpPastFork):
    def __init__(self):
        super().__init__(["Add", "Mul"])


class MoveTransposePastFork(MoveOpPastFork):
    def __init__(self):
        super().__init__(["Transpose"])



# Copyright (c) 2023 Advanced Micro Devices, Inc.
# All rights reserved.
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:
#
# * Redistributions of source code must retain the above copyright notice, this
#   list of conditions and the following disclaimer.
#
# * Redistributions in binary form must reproduce the above copyright notice,
#   this list of conditions and the following disclaimer in the documentation
#   and/or other materials provided with the distribution.
#
# * Neither the name of qonnx nor the names of its
#   contributors may be used to endorse or promote products derived from
#   this software without specific prior written permission.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
# AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
# IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
# DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE
# FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
# DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
# SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER
# CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,
# OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
# OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

import numpy as np
import onnx.parser as oprs

import qonnx.custom_op.general as general
from qonnx.core.modelwrapper import ModelWrapper
from qonnx.custom_op.base import CustomOp
from qonnx.custom_op.registry import getCustomOp


class AttrTestOp(CustomOp):
    def get_nodeattr_types(self):
        my_attrs = {"tensor_attr": ("t", True, np.asarray([])), "strings_attr": ("strings", True, [""])}
        return my_attrs

    def make_shape_compatible_op(self, model):
        param_tensor = self.get_nodeattr("tensor_attr")
        return super().make_const_shape_op(param_tensor.shape)

    def infer_node_datatype(self, model):
        node = self.onnx_node
        # data type stays the same
        dtype = model.get_tensor_datatype(node.input[0])
        model.set_tensor_datatype(node.output[0], dtype)

    def execute_node(self, context, graph):
        node = self.onnx_node
        param_tensor = self.get_nodeattr("tensor_attr")
        context[node.output[0]] = param_tensor

    def verify_node(self):
        pass


def test_attr():
    general.custom_op["AttrTestOp"] = AttrTestOp
    ishp = (1, 10)
    wshp = (1, 3)
    oshp = wshp
    ishp_str = str(list(ishp))
    oshp_str = str(list(oshp))
    wshp_str = str(list(wshp))
    w = np.asarray([1, -2, 3], dtype=np.int8)
    strarr = np.array2string(w, separator=", ")
    w_str = strarr.replace("[", "{").replace("]", "}").replace(" ", "")
    tensor_attr_str = f"int8{wshp_str} {w_str}"
    strings_attr = ["a", "bc", "def"]

    input = f"""
    <
        ir_version: 7,
        opset_import: ["" : 9]
    >
    agraph (float{ishp_str} in0) => (int8{oshp_str} out0)
    {{
        out0 = qonnx.custom_op.general.AttrTestOp<
            tensor_attr={tensor_attr_str}
        >(in0)
    }}
    """
    model = oprs.parse_model(input)
    model = ModelWrapper(model)
    inst = getCustomOp(model.graph.node[0])

    w_prod = inst.get_nodeattr("tensor_attr")
    assert (w_prod == w).all()
    w = w - 1
    inst.set_nodeattr("tensor_attr", w)
    w_prod = inst.get_nodeattr("tensor_attr")
    assert (w_prod == w).all()

    inst.set_nodeattr("strings_attr", strings_attr)
    strings_attr_prod = inst.get_nodeattr("strings_attr")
    assert strings_attr_prod == strings_attr
    strings_attr_prod[0] = "test"
    inst.set_nodeattr("strings_attr", strings_attr_prod)
    assert inst.get_nodeattr("strings_attr") == ["test"] + strings_attr[1:]



# Copyright (c) 2020 Xilinx, Inc.
# All rights reserved.
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:
#
# * Redistributions of source code must retain the above copyright notice, this
#   list of conditions and the following disclaimer.
#
# * Redistributions in binary form must reproduce the above copyright notice,
#   this list of conditions and the following disclaimer in the documentation
#   and/or other materials provided with the distribution.
#
# * Neither the name of Xilinx nor the names of its
#   contributors may be used to endorse or promote products derived from
#   this software without specific prior written permission.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
# AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
# IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
# DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE
# FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
# DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
# SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER
# CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,
# OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
# OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

import numpy as np
import onnx.helper as helper

from qonnx.core.datatype import DataType
from qonnx.custom_op.base import CustomOp


def multithreshold(v, thresholds, out_scale=None, out_bias=None):
    """Given a set of threshold values t={t_0, t_1 ... t_n} the successive
    thresholding maps any real number x to an integer in the interval [0, n],
    where the returned integer is the number of thresholds x is greater than
    or equal to.

    The output tensor will be scaled by out_scale and biased by out_bias."""
    # the inputs are expected to be in the shape (N,C,H,W) or (N, C)
    # the MultiThreshold node supports a data_layout attribute that can be set
    # to 'NHWC' to support (N,H,W,C) data layout mode for in-out as well
    # N : Batch size
    # C : Number of channels
    # H : Heigth of the input images
    # W : Width of the input images
    #
    # the thresholds are expected to be in the shape (C, B)
    # C : Number of channels (must be the same value as C in input tensor
    #     or 1 if all channels use the same threshold value)
    # B : Desired activation steps => i.e. for 4-bit activation,
    #     B=7 (2^(n)-1 and n=4)
    # the output tensor will be scaled by out_scale and biased by out_bias
    # assert threshold shape
    is_global_threshold = thresholds.shape[0] == 1
    assert (
        v.shape[1] == thresholds.shape[0]
    ) or is_global_threshold, """"Threshold
    shape incorrect"""
    # save the required shape sizes for the loops (N, C and B)
    num_batch = v.shape[0]
    num_channel = v.shape[1]
    num_act = thresholds.shape[1]
    # reshape inputs to enable channel-wise reading
    vr = v.reshape((v.shape[0], v.shape[1], -1))
    # initiate output tensor
    ret = np.zeros_like(vr)
    # iterate over thresholds channel-wise
    for t in range(num_channel):
        channel_thresh = thresholds[0] if is_global_threshold else thresholds[t]
        # iterate over batches
        for b in range(num_batch):
            # iterate over the different thresholds for one channel
            for a in range(num_act):
                ret[b][t] += (vr[b][t] >= channel_thresh[a]).astype(int)

    if out_scale is None:
        out_scale = 1.0
    if out_bias is None:
        out_bias = 0.0
    return out_scale * ret.reshape(v.shape) + out_bias


class MultiThreshold(CustomOp):
    """Class that corresponds to a multithresholding node."""

    def get_nodeattr_types(self):
        return {
            "out_dtype": ("s", True, ""),
            "out_scale": ("f", False, 1.0),
            "out_bias": ("f", False, 0.0),
            "data_layout": ("s", False, "NCHW", {"NCHW", "NHWC"}),
        }

    def make_shape_compatible_op(self, model):
        node = self.onnx_node
        return helper.make_node("Relu", [node.input[0]], [node.output[0]])

    def infer_node_datatype(self, model):
        node = self.onnx_node
        odt = self.get_nodeattr("out_dtype")
        is_float = False
        scale = self.get_nodeattr("out_scale")
        bias = self.get_nodeattr("out_bias")
        if scale is not None and (int(scale) != scale):
            is_float = True
        if bias is not None and (int(bias) != bias):
            is_float = True
        if is_float:
            model.set_tensor_datatype(node.output[0], DataType["FLOAT32"])
        else:
            model.set_tensor_datatype(node.output[0], DataType[odt])

    def execute_node(self, context, graph):
        node = self.onnx_node
        # save inputs
        v = context[node.input[0]]
        thresholds = context[node.input[1]]
        # retrieve attributes if output scaling is used
        out_scale = self.get_nodeattr("out_scale")
        out_bias = self.get_nodeattr("out_bias")
        # transpose input if NHWC data layout is chosen
        data_layout = self.get_nodeattr("data_layout")
        if data_layout == "NHWC":
            if v.ndim == 4:
                # NHWC -> NCHW
                v = np.transpose(v, (0, 3, 1, 2))
            elif v.ndim == 2:
                # no HW dimension means NHWC and NCHW layouts are equivalent
                pass
            else:
                raise Exception("Unknown data_layout and input ndim" " combination for MultiThreshold.")

        # Remember whether the shape has been modified to handle 1d or 3d data
        # layouts
        orig_shape = None
        # If the input tensor has dimensions not covered by the NC or NCWH data
        # layouts, the shape needs to be adapted such that it can be handled by
        # multithreshold.
        # TODO: Seems like a rather sketchy solution to support arbitrary data
        #  layouts. This does not even validate the assumption of channel last
        #  layout.
        if v.ndim not in {2, 4}:
            # Remember the original shape to be restored later
            orig_shape = v.shape
            # Assume last dimension to be the channel dimension C and reshape
            # into NC layout which is supported by multithreshold
            v = v.reshape((-1, v.shape[-1]))

        # calculate output
        output = multithreshold(v, thresholds, out_scale, out_bias)
        # setting context according to output
        if data_layout == "NHWC":
            if output.ndim == 4:
                # NCHW -> NHWC
                output = np.transpose(output, (0, 2, 3, 1))
            elif output.ndim == 2:
                # no HW dimension means NHWC and NCHW layouts are equivalent
                pass
            else:
                raise Exception("Unknown data_layout and output ndim" " combination for MultiThreshold.")

        # If the shape has been modified to support arbitrary layouts, restore
        # the original shape
        # TODO: Part of the rather sketchy solution above.
        if orig_shape is not None:
            output = output.reshape(orig_shape)

        context[node.output[0]] = output

    def verify_node(self):
        info_messages = []

        # verify that all necessary attributes exist
        try:
            self.get_nodeattr("out_dtype")
            info_messages.append("All necessary attributes exist")
        except Exception:
            info_messages.append(
                """The necessary attributes do not exist.
                MultiThreshold needs the following attributes:
                out_scale, out_bias, out_dtype"""
            )

        # verify the number of inputs
        if len(self.onnx_node.input) == 2:
            info_messages.append("The number of inputs is correct")
        else:
            info_messages.append(
                """MultiThreshold needs 2 inputs
                    (data input and threshold values)"""
            )

        return info_messages



# Copyright (c) 2023 Advanced Micro Devices, Inc.
# All rights reserved.
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:
#
# * Redistributions of source code must retain the above copyright notice, this
#   list of conditions and the following disclaimer.
#
# * Redistributions in binary form must reproduce the above copyright notice,
#   this list of conditions and the following disclaimer in the documentation
#   and/or other materials provided with the distribution.
#
# * Neither the name of qonnx nor the names of its
#   contributors may be used to endorse or promote products derived from
#   this software without specific prior written permission.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
# AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
# IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
# DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE
# FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
# DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
# SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER
# CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,
# OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
# OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

import pytest

import numpy as np

from qonnx.util.range_analysis import range_analysis
from qonnx.util.test import download_model, test_model_details

model_details_stuckchans = {
    "MobileNetv1-w4a4": {
        "stuck_chans": {
            "Quant_29_out0": [
                (0, 0.4813263),
                (4, 0.0),
                (6, 0.0),
                (10, 0.0),
                (13, 0.0),
                (15, 0.0),
                (16, 0.0),
                (19, 0.0),
                (26, 0.0),
                (28, 0.0),
            ],
            "Quant_30_out0": [
                (0, 0.0),
                (4, 0.0),
                (6, 0.0),
                (10, 0.0),
                (13, 0.15743902),
                (15, 0.0),
                (16, 0.47231707),
                (19, 0.0),
                (26, 0.0),
                (28, 0.0),
            ],
            "Quant_31_out0": [(42, 0.0)],
            "Quant_32_out0": [(42, 0.0)],
            "Quant_35_out0": [(102, 0.0)],
            "Quant_36_out0": [(102, 0.0)],
        }
    },
    "FINN-CNV_W2A2": {
        "stuck_chans": {
            "Quant_10_out0": [(5, -1.0), (10, 1.0), (26, 1.0), (30, -1.0), (34, -1.0), (54, -1.0)],
            "Quant_11_out0": [(30, 1.0), (35, 1.0), (37, -1.0), (42, 1.0), (45, -1.0), (57, -1.0)],
            "Quant_13_out0": [(40, -1.0)],
            "Quant_14_out0": [(4, 1.0), (175, 1.0), (209, -1.0)],
            "Quant_16_out0": [
                (5, -1.0),
                (50, 1.0),
                (77, -1.0),
                (95, -1.0),
                (153, 1.0),
                (186, 1.0),
                (199, 1.0),
                (209, -1.0),
                (241, 1.0),
                (329, 1.0),
                (340, 1.0),
                (465, -1.0),
                (478, -1.0),
                (510, -1.0),
            ],
            "Quant_17_out0": [(101, -0.0), (230, -0.0), (443, 0.0)],
        }
    },
}

# inherit basics for matching testcases from test util
model_details = {k: v for (k, v) in test_model_details.items() if k in model_details_stuckchans.keys()}
model_details = {**model_details, **model_details_stuckchans}


@pytest.mark.parametrize("model_name", model_details.keys())
def test_range_analysis(model_name):
    model = download_model(model_name, return_modelwrapper=True)
    irange = test_model_details[model_name]["input_range"]
    ret = range_analysis(model, irange=irange, report_mode="stuck_channel", key_filter="Quant", do_cleanup=True)
    golden_stuck_channels = model_details[model_name]["stuck_chans"]
    for tname, ret_chans in ret.items():
        tg_chans = golden_stuck_channels[tname]
        for i in range(len(tg_chans)):
            tg_ind, tg_val = tg_chans[i]
            ret_ind, ret_val = ret_chans[i]
            assert tg_ind == ret_ind
            assert np.isclose(tg_val, ret_val)



# Copyright (c) 2020 Xilinx, Inc.
# All rights reserved.
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:
#
# * Redistributions of source code must retain the above copyright notice, this
#   list of conditions and the following disclaimer.
#
# * Redistributions in binary form must reproduce the above copyright notice,
#   this list of conditions and the following disclaimer in the documentation
#   and/or other materials provided with the distribution.
#
# * Neither the name of Xilinx nor the names of its
#   contributors may be used to endorse or promote products derived from
#   this software without specific prior written permission.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
# AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
# IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
# DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE
# FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
# DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
# SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER
# CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,
# OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
# OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

import json
import numpy as np
import onnx
import onnx.parser as oprs
import os
from pkgutil import get_data

import qonnx.core.onnx_exec as oxe
from qonnx.core.datatype import DataType
from qonnx.core.modelwrapper import ModelWrapper
from qonnx.custom_op.registry import getCustomOp
from qonnx.transformation.general import ApplyConfig, ConvertDivToMul, GiveUniqueNodeNames, GiveUniqueParameterTensors
from qonnx.transformation.infer_shapes import InferShapes
from qonnx.transformation.lower_convs_to_matmul import LowerConvsToMatMul
from qonnx.util.basic import gen_finn_dt_tensor, qonnx_make_model


def test_mul_to_div():
    shp = (1, 2, 4)
    dt0 = DataType["UINT8"]
    np.random.seed(0)
    shp_str = str(list(shp))

    input = f"""
    <
        ir_version: 7,
        opset_import: ["" : 9]
    >
    agraph (float{shp_str} in0) => (float{shp_str} out0)
    <
        float param_div = {{2.0}}
    >
    {{
        out0 = Div(in0, param_div)
    }}
    """
    model = oprs.parse_model(input)
    model = ModelWrapper(model)
    model.set_tensor_datatype("in0", dt0)
    model = model.transform(InferShapes())
    inp = gen_finn_dt_tensor(dt0, shp)
    input_dict = {"in0": inp}
    out_expected = oxe.execute_onnx(model, input_dict)["out0"]
    new_model = model.transform(ConvertDivToMul())
    out_produced = oxe.execute_onnx(new_model, input_dict)["out0"]
    assert (out_expected == out_produced).all()


def test_give_unique_node_names():
    raw_m = get_data("qonnx.data", "onnx/mnist-conv/model.onnx")
    model = ModelWrapper(raw_m)
    model = model.transform(GiveUniqueNodeNames())
    assert model.graph.node[0].name == "Reshape_0"
    assert model.graph.node[1].name == "Conv_0"
    assert model.graph.node[11].name == "Add_2"


def test_give_unique_node_names_missingshape():
    # see https://github.com/fastmachinelearning/qonnx/issues/33
    Add1_node = onnx.helper.make_node("Add", inputs=["in1", "Add_1_param0"], outputs=["sum1"], name="Add_1")

    Add0_node = onnx.helper.make_node(
        "Add",
        inputs=["sum1", "Add_0_param0"],
        outputs=["sum2"],
        name="Add_0",
    )

    Add2_node = onnx.helper.make_node(
        "Add",
        inputs=["abs1", "abs1"],
        outputs=["sum3"],
        name="Add_2",
    )

    Abs_node = onnx.helper.make_node("Abs", inputs=["sum2"], outputs=["abs1"], name="Abs")

    Round_node = onnx.helper.make_node(
        "Round",
        inputs=["sum3"],
        outputs=["out1"],
        name="Round",
    )

    in1 = onnx.helper.make_tensor_value_info("in1", onnx.TensorProto.FLOAT, [4, 4])
    out1 = onnx.helper.make_tensor_value_info("out1", onnx.TensorProto.FLOAT, [4, 4])

    graph = onnx.helper.make_graph(
        nodes=[
            Add1_node,
            Add0_node,
            Abs_node,
            Add2_node,
            Round_node,
        ],
        name="simple_graph",
        inputs=[in1],
        outputs=[out1],
        value_info=[
            # note: value_info for sum1 explicitly omitted to trigger bug #33
            # onnx.helper.make_tensor_value_info("sum1", onnx.TensorProto.FLOAT, [4, 4]),
            onnx.helper.make_tensor_value_info("sum2", onnx.TensorProto.FLOAT, [4, 4]),
            onnx.helper.make_tensor_value_info("abs1", onnx.TensorProto.FLOAT, [4, 4]),
            onnx.helper.make_tensor_value_info("sum3", onnx.TensorProto.FLOAT, [4, 4]),
        ],
        initializer=[
            onnx.helper.make_tensor("Add_1_param0", onnx.TensorProto.FLOAT, [4, 4], np.zeros(16).tolist()),
            onnx.helper.make_tensor("Add_0_param0", onnx.TensorProto.FLOAT, [4, 4], np.zeros(16).tolist()),
        ],
    )

    onnx_model = onnx.helper.make_model(graph, producer_name="simple-model")
    model = ModelWrapper(onnx_model)
    # sum1 is missing shape so should return False
    assert model.check_all_tensor_shapes_specified() is False
    model = model.transform(GiveUniqueNodeNames())
    assert [x.name for x in model.get_nodes_by_op_type("Add")] == ["Add_0", "Add_1", "Add_2"]
    assert [x.name for x in model.get_nodes_by_op_type("Abs")] == ["Abs_0"]
    assert [x.name for x in model.get_nodes_by_op_type("Round")] == ["Round_0"]


def test_give_unique_parameter_tensors():
    # Create model
    input_shape = [4, 4]
    in1 = onnx.helper.make_tensor_value_info("in1", onnx.TensorProto.FLOAT, input_shape)
    out1 = onnx.helper.make_tensor_value_info("out1", onnx.TensorProto.FLOAT, input_shape)

    graph_nodes = []
    graph_nodes += [onnx.helper.make_node("Add", inputs=["in1", "param1"], outputs=["t1"])]

    graph_nodes += [onnx.helper.make_node("Sum", inputs=["t1", "param1", "param1"], outputs=["t2"])]

    graph_nodes += [onnx.helper.make_node("Sum", inputs=["t2", "param2", "param1"], outputs=["t3"])]

    graph_nodes += [onnx.helper.make_node("Add", inputs=["t3", "param1"], outputs=["out1"])]

    onnx_graph = onnx.helper.make_graph(
        nodes=graph_nodes,
        name="simple_graph",
        inputs=[in1],
        outputs=[out1],
    )

    onnx_model = qonnx_make_model(onnx_graph, producer_name="simple-model")
    model = ModelWrapper(onnx_model)

    # Set param values
    np.random.seed(0)
    param1 = np.random.rand(*input_shape).astype(np.float32)
    param2 = np.random.rand(*input_shape).astype(np.float32)
    model.set_initializer("param1", param1)
    model.set_initializer("param2", param2)
    model = model.transform(InferShapes())

    # Apply transformation
    new_model = model.transform(GiveUniqueParameterTensors())
    new_model = new_model.transform(InferShapes())

    # Test
    # Breaks the model?
    input_tensor = np.random.rand(*input_shape).astype(np.float32)
    input_dict = {"in1": input_tensor}

    # run original
    expected_context = oxe.execute_onnx(model, input_dict)
    expected_output = expected_context[model.graph.output[0].name]

    # run modified
    produced_context = oxe.execute_onnx(new_model, input_dict)
    produced_output = produced_context[new_model.graph.output[0].name]

    assert np.isclose(
        expected_output, produced_output, atol=1e-8
    ).all(), " GiveUniqueParameterTensors() transform breaks the model"

    # Does the job?
    param_set = set()
    param_cnt = 0
    for n in new_model.graph.node:
        for i in range(1, len(n.input)):
            param_set |= {n.input[i]}
            param_cnt += 1

    assert len(param_set) == param_cnt, " There are still parameters reused"


def test_apply_config():
    raw_m = get_data("qonnx.data", "onnx/mnist-conv/model.onnx")
    model = ModelWrapper(raw_m)
    model = model.transform(GiveUniqueNodeNames())
    model = model.transform(LowerConvsToMatMul())
    model = model.transform(GiveUniqueNodeNames())
    # set up a config in a dict, then dump it to JSON
    config = {}
    config["Defaults"] = {"kernel_size": [[3, 3], ["Im2Col"]]}
    config["Im2Col_0"] = {"kernel_size": [7, 7]}
    with open("config.json", "w") as f:
        json.dump(config, f, indent=4)
    model = model.transform(ApplyConfig("config.json"))
    # check model
    assert getCustomOp(model.graph.node[2]).get_nodeattr("kernel_size") == [7, 7]
    assert getCustomOp(model.graph.node[9]).get_nodeattr("kernel_size") == [3, 3]
    os.remove("config.json")



# Copyright (c) 2021 Xilinx, Inc.
# All rights reserved.
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:
#
# * Redistributions of source code must retain the above copyright notice, this
#   list of conditions and the following disclaimer.
#
# * Redistributions in binary form must reproduce the above copyright notice,
#   this list of conditions and the following disclaimer in the documentation
#   and/or other materials provided with the distribution.
#
# * Neither the name of Xilinx nor the names of its
#   contributors may be used to endorse or promote products derived from
#   this software without specific prior written permission.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
# AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
# IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
# DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE
# FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
# DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
# SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER
# CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,
# OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
# OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

from onnx import helper as oh

import qonnx.core.data_layout as data_layout
from qonnx.transformation.base import Transformation


class MakeInputChannelsLast(Transformation):
    """For networks with an input using the NCx data layout, add a transpose node
    at the beginning and mark the input as using NxC (channels-last)."""

    def __init__(self):
        super().__init__()

    def apply(self, model):
        graph_in_name = model.graph.input[0].name
        graph_new_in_name = graph_in_name + "_transposed"
        orig_ishape = model.get_tensor_shape(graph_in_name)
        ndim = len(orig_ishape)
        if ndim == 2:
            # assume NC layout, no action needed
            return (model, False)
        elif ndim > 2:
            orig_layout = model.get_tensor_layout(graph_in_name)
            if orig_layout == data_layout.get_channels_last_layout_for_ndims(ndim):
                # already marked as channels-last, no action needed
                return (model, False)
            else:
                # determine channels-last shape and required permutation to
                # go from channels-last to previous format
                new_perm = list(range(ndim))
                new_perm.remove(ndim - 1)
                new_perm.insert(1, ndim - 1)
                new_ishape = list(orig_ishape)
                new_ishape.remove(orig_ishape[1])
                new_ishape.append(orig_ishape[1])
                # create and insert transpose node
                t_trans_node = oh.make_node("Transpose", [graph_in_name], [graph_new_in_name], perm=new_perm)
                model.graph.node.insert(0, t_trans_node)
                # rewire all consumers of original input to transpose's output
                consumers = model.find_consumers(graph_in_name)
                for cons in consumers:
                    if cons == t_trans_node:
                        continue
                    for i, ci in enumerate(cons.input):
                        if ci == graph_in_name:
                            cons.input[i] = graph_new_in_name
                # set tensor shapes and layouts
                model.set_tensor_shape(graph_in_name, new_ishape)
                model.set_tensor_shape(graph_new_in_name, orig_ishape)
                model.set_tensor_layout(graph_in_name, data_layout.get_channels_last_layout_for_ndims(ndim))
                model.set_tensor_layout(
                    graph_new_in_name,
                    data_layout.get_channels_first_layout_for_ndims(ndim),
                )
                # single iteration is enough so return model_was_changed=False
                return (model, False)
